import datetime
import json
from typing import Any, Dict
import proxai.types as types
import proxai.stat_types as stat_types
import proxai.connectors.model_configs as model_configs


def encode_provider_model_type(
    provider_model_type: types.ProviderModelType) -> Dict[str, Any]:
  record = {}
  record['provider'] = provider_model_type.provider
  record['model'] = provider_model_type.model
  record['provider_model_identifier'] = (
      provider_model_type.provider_model_identifier)
  return record


def decode_provider_model_type(
    record: Dict[str, Any]) -> types.ProviderModelType:
  if 'provider' not in record:
    raise ValueError(f'Provider not found in record: {record=}')
  if 'model' not in record:
    raise ValueError(f'Model not found in record: {record=}')
  if 'provider_model_identifier' not in record:
    raise ValueError(
        f'Provider model identifier not found in record: {record=}')
  provider_model = model_configs.ALL_MODELS[record['provider']][record['model']]
  if provider_model.provider_model_identifier != record[
      'provider_model_identifier']:
    raise ValueError(
        'Provider model identifier mismatch: '
        f'{record["provider_model_identifier"]} != '
        f'{provider_model.provider_model_identifier}'
        '\nThis can happen if the model config has changed in recent versions.')
  return provider_model


def encode_query_record(
    query_record: types.QueryRecord) -> Dict[str, Any]:
  record = {}
  if query_record.call_type != None:
    record['call_type'] = query_record.call_type.value
  if query_record.provider_model != None:
    record['provider_model'] = encode_provider_model_type(
        query_record.provider_model)
  if query_record.prompt != None:
    record['prompt'] = query_record.prompt
  if query_record.system != None:
    record['system'] = query_record.system
  if query_record.messages != None:
    record['messages'] = query_record.messages
  if query_record.max_tokens != None:
    record['max_tokens'] = str(query_record.max_tokens)
  if query_record.temperature != None:
    record['temperature'] = str(query_record.temperature)
  if query_record.stop != None:
    record['stop'] = query_record.stop
  if query_record.hash_value != None:
    record['hash_value'] = query_record.hash_value
  if query_record.token_count != None:
    record['token_count'] = str(query_record.token_count)
  return record


def decode_query_record(
    record: Dict[str, Any]) -> types.QueryRecord:
  query_record = types.QueryRecord()
  if 'call_type' in record:
    query_record.call_type = types.CallType(record['call_type'])
  if 'provider_model' in record:
    query_record.provider_model = decode_provider_model_type(
        record['provider_model'])
  query_record.prompt = record.get('prompt', None)
  query_record.system = record.get('system', None)
  query_record.messages = record.get('messages', None)
  if 'max_tokens' in record:
    query_record.max_tokens = int(record['max_tokens'])
  if 'temperature' in record:
    query_record.temperature = float(record['temperature'])
  query_record.stop = record.get('stop', None)
  query_record.hash_value = record.get('hash_value', None)
  if 'token_count' in record:
    query_record.token_count = int(record['token_count'])
  return query_record


def encode_query_response_record(
    query_response_record: types.QueryResponseRecord
) -> Dict[str, Any]:
  record = {}
  if query_response_record.response != None:
    record['response'] = query_response_record.response
  if query_response_record.error != None:
    record['error'] = query_response_record.error
  if query_response_record.start_utc_date != None:
    record['start_utc_date'] = query_response_record.start_utc_date.isoformat()
  if query_response_record.end_utc_date != None:
    record['end_utc_date'] = query_response_record.end_utc_date.isoformat()
  if query_response_record.local_time_offset_minute != None:
    record['local_time_offset_minute'] = (
        query_response_record.local_time_offset_minute)
  if query_response_record.response_time != None:
    record['response_time'] = (
        query_response_record.response_time.total_seconds())
  if query_response_record.estimated_cost != None:
    record['estimated_cost'] = query_response_record.estimated_cost
  if query_response_record.token_count != None:
    record['token_count'] = str(query_response_record.token_count)
  return record


def decode_query_response_record(
    record: Dict[str, Any]) -> types.QueryResponseRecord:
  query_response_record = types.QueryResponseRecord()
  query_response_record.response = record.get('response', None)
  query_response_record.error = record.get('error', None)
  if 'start_utc_date' in record:
    query_response_record.start_utc_date = datetime.datetime.fromisoformat(
        record['start_utc_date'])
  if 'end_utc_date' in record:
    query_response_record.end_utc_date = datetime.datetime.fromisoformat(
        record['end_utc_date'])
  if 'local_time_offset_minute' in record:
    query_response_record.local_time_offset_minute = (
        record['local_time_offset_minute'])
  if 'response_time' in record:
    query_response_record.response_time = datetime.timedelta(
        seconds=record['response_time'])
  if 'estimated_cost' in record:
    query_response_record.estimated_cost = record['estimated_cost']
  if 'token_count' in record:
    query_response_record.token_count = int(record['token_count'])
  return query_response_record


def encode_cache_record(
    cache_record: types.CacheRecord) -> Dict[str, Any]:
  record = {}
  if cache_record.query_record != None:
    record['query_record'] = encode_query_record(
        cache_record.query_record)
  if cache_record.query_responses != None:
    record['query_responses'] = []
    for query_response_record in cache_record.query_responses:
      record['query_responses'].append(
          encode_query_response_record(query_response_record))
  if cache_record.shard_id != None:
    try:
      record['shard_id'] = int(cache_record.shard_id)
    except ValueError:
      record['shard_id'] = cache_record.shard_id
  if cache_record.last_access_time != None:
    record['last_access_time'] = cache_record.last_access_time.isoformat()
  if cache_record.call_count != None:
    record['call_count'] = cache_record.call_count
  return record


def decode_cache_record(
    record: Dict[str, Any]) -> types.CacheRecord:
  cache_record = types.CacheRecord()
  if 'query_record' in record:
    cache_record.query_record = decode_query_record(
        record['query_record'])
  if 'query_responses' in record:
    cache_record.query_responses = []
    for query_response_record in record['query_responses']:
      cache_record.query_responses.append(
          decode_query_response_record(query_response_record))
  if 'shard_id' in record:
    try:
      cache_record.shard_id = int(record['shard_id'])
    except ValueError:
      cache_record.shard_id = record['shard_id']
  if 'last_access_time' in record:
    cache_record.last_access_time = datetime.datetime.fromisoformat(
        record['last_access_time'])
  if 'call_count' in record:
    cache_record.call_count = int(record['call_count'])
  return cache_record


def encode_light_cache_record(
    light_cache_record: types.LightCacheRecord) -> Dict[str, Any]:
  record = {}
  if light_cache_record.query_record_hash != None:
    record['query_record_hash'] = light_cache_record.query_record_hash
  if light_cache_record.query_response_count != None:
    record['query_response_count'] = light_cache_record.query_response_count
  if light_cache_record.shard_id != None:
    try:
      record['shard_id'] = int(light_cache_record.shard_id)
    except ValueError:
      record['shard_id'] = light_cache_record.shard_id
  if light_cache_record.last_access_time != None:
    record['last_access_time'] = (
        light_cache_record.last_access_time.isoformat())
  if light_cache_record.call_count != None:
    record['call_count'] = light_cache_record.call_count
  return record


def decode_light_cache_record(
    record: Dict[str, Any]) -> types.LightCacheRecord:
  light_cache_record = types.LightCacheRecord()
  light_cache_record.query_record_hash = record.get('query_record_hash', None)
  if 'query_response_count' in record:
    light_cache_record.query_response_count = int(
        record['query_response_count'])
  if 'shard_id' in record:
    try:
      light_cache_record.shard_id = int(record['shard_id'])
    except ValueError:
      light_cache_record.shard_id = record['shard_id']
  if 'last_access_time' in record:
    light_cache_record.last_access_time = datetime.datetime.fromisoformat(
        record['last_access_time'])
  if 'call_count' in record:
    light_cache_record.call_count = int(record['call_count'])
  return light_cache_record


def encode_logging_record(
    logging_record: types.LoggingRecord) -> Dict[str, Any]:
  record = {}
  if logging_record.query_record != None:
    record['query_record'] = encode_query_record(
        logging_record.query_record)
  if logging_record.response_record != None:
    record['response_record'] = encode_query_response_record(
        logging_record.response_record)
  if logging_record.response_source != None:
    record['response_source'] = logging_record.response_source.value
  if logging_record.look_fail_reason != None:
    record['look_fail_reason'] = logging_record.look_fail_reason.value
  return record


def decode_logging_record(
    record: Dict[str, Any]) -> types.LoggingRecord:
  logging_record = types.LoggingRecord()
  if 'query_record' in record:
    logging_record.query_record = decode_query_record(
        record['query_record'])
  if 'response_record' in record:
    logging_record.response_record = decode_query_response_record(
        record['response_record'])
  if 'response_source' in record:
    logging_record.response_source = (
        types.ResponseSource(record['response_source']))
  if 'look_fail_reason' in record:
    logging_record.look_fail_reason = (
        types.CacheLookFailReason(record['look_fail_reason']))
  return logging_record


def encode_model_status(
    model_status: types.ModelStatus) -> Dict[str, Any]:
  record = {}
  if model_status.unprocessed_models != None:
    record['unprocessed_models'] = []
    for provider_model in model_status.unprocessed_models:
      record['unprocessed_models'].append(
          encode_provider_model_type(provider_model))
  if model_status.working_models != None:
    record['working_models'] = []
    for provider_model in model_status.working_models:
      record['working_models'].append(
          encode_provider_model_type(provider_model))
  if model_status.failed_models != None:
    record['failed_models'] = []
    for provider_model in model_status.failed_models:
      record['failed_models'].append(encode_provider_model_type(provider_model))
  if model_status.filtered_models != None:
    record['filtered_models'] = []
    for provider_model in model_status.filtered_models:
      record['filtered_models'].append(
          encode_provider_model_type(provider_model))
  if model_status.provider_queries != None:
    record['provider_queries'] = {}
    for provider_model, provider_query in model_status.provider_queries.items():
      provider_model = json.dumps(encode_provider_model_type(provider_model))
      record['provider_queries'][provider_model] = (
          encode_logging_record(provider_query))
  return record


def decode_model_status(
    record: Dict[str, Any]) -> types.ModelStatus:
  model_status = types.ModelStatus()
  if 'unprocessed_models' in record:
    for provider_model_record in record['unprocessed_models']:
      model_status.unprocessed_models.add(
          decode_provider_model_type(provider_model_record))
  if 'working_models' in record:
    for provider_model_record in record['working_models']:
      model_status.working_models.add(
          decode_provider_model_type(provider_model_record))
  if 'failed_models' in record:
    for provider_model_record in record['failed_models']:
      model_status.failed_models.add(
          decode_provider_model_type(provider_model_record))
  if 'filtered_models' in record:
    for provider_model_record in record['filtered_models']:
      model_status.filtered_models.add(
          decode_provider_model_type(provider_model_record))
  if 'provider_queries' in record:
    for provider_model, provider_query_record in record[
        'provider_queries'].items():
      provider_model = json.loads(provider_model)
      provider_model = decode_provider_model_type(provider_model)
      model_status.provider_queries[provider_model] = (
          decode_logging_record(provider_query_record))
  return model_status


def encode_logging_options(
    logging_options: types.LoggingOptions) -> Dict[str, Any]:
  record = {}
  if logging_options.logging_path != None:
    record['logging_path'] = logging_options.logging_path
  if logging_options.stdout != None:
    record['stdout'] = logging_options.stdout
  if logging_options.hide_sensitive_content != None:
    record['hide_sensitive_content'] = logging_options.hide_sensitive_content
  return record


def encode_cache_options(
    cache_options: types.CacheOptions) -> Dict[str, Any]:
  record = {}
  if cache_options.cache_path != None:
    record['cache_path'] = cache_options.cache_path
  if cache_options.unique_response_limit != None:
    record['unique_response_limit'] = cache_options.unique_response_limit
  if cache_options.retry_if_error_cached != None:
    record['retry_if_error_cached'] = cache_options.retry_if_error_cached
  if cache_options.clear_query_cache_on_connect != None:
    record['clear_query_cache_on_connect'] = (
        cache_options.clear_query_cache_on_connect)
  if cache_options.clear_model_cache_on_connect != None:
    record['clear_model_cache_on_connect'] = (
        cache_options.clear_model_cache_on_connect)
  return record


def encode_proxdash_options(
    proxdash_options: types.ProxDashOptions) -> Dict[str, Any]:
  record = {}
  if proxdash_options.stdout != None:
    record['stdout'] = proxdash_options.stdout
  if proxdash_options.hide_sensitive_content != None:
    record['hide_sensitive_content'] = proxdash_options.hide_sensitive_content
  if proxdash_options.disable_proxdash != None:
    record['disable_proxdash'] = proxdash_options.disable_proxdash
  return record


def encode_run_options(
    run_options: types.RunOptions) -> Dict[str, Any]:
  record = {}
  if run_options.run_type != None:
    record['run_type'] = run_options.run_type.value
  if run_options.hidden_run_key != None:
    record['hidden_run_key'] = run_options.hidden_run_key
  if run_options.experiment_path != None:
    record['experiment_path'] = run_options.experiment_path
  if run_options.root_logging_path != None:
    record['root_logging_path'] = run_options.root_logging_path
  if run_options.default_model_cache_path != None:
    record['default_model_cache_path'] = run_options.default_model_cache_path
  if run_options.logging_options != None:
    record['logging_options'] = encode_logging_options(
        run_options.logging_options)
  if run_options.cache_options != None:
    record['cache_options'] = encode_cache_options(
        run_options.cache_options)
  if run_options.proxdash_options != None:
    record['proxdash_options'] = encode_proxdash_options(
        run_options.proxdash_options)
  if run_options.allow_multiprocessing != None:
    record['allow_multiprocessing'] = run_options.allow_multiprocessing
  if run_options.model_test_timeout != None:
    record['model_test_timeout'] = run_options.model_test_timeout
  if run_options.strict_feature_test != None:
    record['strict_feature_test'] = run_options.strict_feature_test
  if run_options.suppress_provider_errors != None:
    record['suppress_provider_errors'] = run_options.suppress_provider_errors
  return record


def decode_logging_options(
    record: Dict[str, Any]) -> types.LoggingOptions:
  logging_options = types.LoggingOptions()
  if 'logging_path' in record:
    logging_options.logging_path = record['logging_path']
  if 'stdout' in record:
    logging_options.stdout = record['stdout']
  if 'hide_sensitive_content' in record:
    logging_options.hide_sensitive_content = record['hide_sensitive_content']
  return logging_options


def decode_cache_options(
    record: Dict[str, Any]) -> types.CacheOptions:
  cache_options = types.CacheOptions()
  if 'cache_path' in record:
    cache_options.cache_path = record['cache_path']
  if 'unique_response_limit' in record:
    cache_options.unique_response_limit = record['unique_response_limit']
  if 'retry_if_error_cached' in record:
    cache_options.retry_if_error_cached = record['retry_if_error_cached']
  if 'clear_query_cache_on_connect' in record:
    cache_options.clear_query_cache_on_connect = (
        record['clear_query_cache_on_connect'])
  if 'clear_model_cache_on_connect' in record:
    cache_options.clear_model_cache_on_connect = (
        record['clear_model_cache_on_connect'])
  return cache_options


def decode_proxdash_options(
    record: Dict[str, Any]) -> types.ProxDashOptions:
  proxdash_options = types.ProxDashOptions()
  if 'stdout' in record:
    proxdash_options.stdout = record['stdout']
  if 'hide_sensitive_content' in record:
    proxdash_options.hide_sensitive_content = record['hide_sensitive_content']
  if 'disable_proxdash' in record:
    proxdash_options.disable_proxdash = record['disable_proxdash']
  return proxdash_options


def decode_run_options(
    record: Dict[str, Any]) -> types.RunOptions:
  run_options = types.RunOptions()
  if 'run_type' in record:
    run_options.run_type = types.RunType(record['run_type'])
  if 'hidden_run_key' in record:
    run_options.hidden_run_key = record['hidden_run_key']
  if 'experiment_path' in record:
    run_options.experiment_path = record['experiment_path']
  if 'root_logging_path' in record:
    run_options.root_logging_path = record['root_logging_path']
  if 'default_model_cache_path' in record:
    run_options.default_model_cache_path = record['default_model_cache_path']
  if 'logging_options' in record:
    run_options.logging_options = decode_logging_options(
        record['logging_options'])
  if 'cache_options' in record:
    run_options.cache_options = decode_cache_options(
        record['cache_options'])
  if 'proxdash_options' in record:
    run_options.proxdash_options = decode_proxdash_options(
        record['proxdash_options'])
  if 'allow_multiprocessing' in record:
    run_options.allow_multiprocessing = record['allow_multiprocessing']
  if 'model_test_timeout' in record:
    run_options.model_test_timeout = record['model_test_timeout']
  if 'strict_feature_test' in record:
    run_options.strict_feature_test = record['strict_feature_test']
  if 'suppress_provider_errors' in record:
    run_options.suppress_provider_errors = record['suppress_provider_errors']
  return run_options


def encode_base_provider_stats(
    base_provider_stats: stat_types.BaseProviderStats) -> Dict[str, Any]:
  record = {}
  if base_provider_stats.total_queries:
    record['total_queries'] = base_provider_stats.total_queries
  if base_provider_stats.total_successes:
    record['total_successes'] = base_provider_stats.total_successes
  if base_provider_stats.total_fails:
    record['total_fails'] = base_provider_stats.total_fails
  if base_provider_stats.total_token_count:
    record['total_token_count'] = base_provider_stats.total_token_count
  if base_provider_stats.total_query_token_count:
    record['total_query_token_count'] = (
        base_provider_stats.total_query_token_count)
  if base_provider_stats.total_response_token_count:
    record['total_response_token_count'] = (
        base_provider_stats.total_response_token_count)
  if base_provider_stats.total_response_time:
    record['total_response_time'] = base_provider_stats.total_response_time
  if base_provider_stats.avr_response_time:
    record['avr_response_time'] = base_provider_stats.avr_response_time
  if base_provider_stats.estimated_cost:
    record['estimated_cost'] = base_provider_stats.estimated_cost
  if base_provider_stats.total_cache_look_fail_reasons:
    record['total_cache_look_fail_reasons'] = {}
    for k, v in base_provider_stats.total_cache_look_fail_reasons.items():
      record['total_cache_look_fail_reasons'][k.value] = v
  return record


def decode_base_provider_stats(
    record: Dict[str, Any]) -> stat_types.BaseProviderStats:
  base_provider_stats = stat_types.BaseProviderStats()
  if 'total_queries' in record:
    base_provider_stats.total_queries = record['total_queries']
  if 'total_successes' in record:
    base_provider_stats.total_successes = record['total_successes']
  if 'total_fails' in record:
    base_provider_stats.total_fails = record['total_fails']
  if 'total_token_count' in record:
    base_provider_stats.total_token_count = record['total_token_count']
  if 'total_query_token_count' in record:
    base_provider_stats.total_query_token_count = (
        record['total_query_token_count'])
  if 'total_response_token_count' in record:
    base_provider_stats.total_response_token_count = (
        record['total_response_token_count'])
  if 'total_response_time' in record:
    base_provider_stats.total_response_time = record['total_response_time']
  if 'estimated_cost' in record:
    base_provider_stats.estimated_cost = record['estimated_cost']
  if 'total_cache_look_fail_reasons' in record:
    base_provider_stats.total_cache_look_fail_reasons = {}
    for k, v in record['total_cache_look_fail_reasons'].items():
      base_provider_stats.total_cache_look_fail_reasons[
          types.CacheLookFailReason(k)] = v
  return base_provider_stats


def encode_base_cache_stats(
    base_cache_stats: stat_types.BaseCacheStats) -> Dict[str, Any]:
  record = {}
  if base_cache_stats.total_cache_hit:
    record['total_cache_hit'] = base_cache_stats.total_cache_hit
  if base_cache_stats.total_success_return:
    record['total_success_return'] = base_cache_stats.total_success_return
  if base_cache_stats.total_fail_return:
    record['total_fail_return'] = base_cache_stats.total_fail_return
  if base_cache_stats.saved_token_count:
    record['saved_token_count'] = base_cache_stats.saved_token_count
  if base_cache_stats.saved_query_token_count:
    record['saved_query_token_count'] = base_cache_stats.saved_query_token_count
  if base_cache_stats.saved_response_token_count:
    record['saved_response_token_count'] = (
        base_cache_stats.saved_response_token_count)
  if base_cache_stats.saved_total_response_time:
    record['saved_total_response_time'] = (
        base_cache_stats.saved_total_response_time)
  if base_cache_stats.saved_avr_response_time:
    record['saved_avr_response_time'] = base_cache_stats.saved_avr_response_time
  if base_cache_stats.saved_estimated_cost:
    record['saved_estimated_cost'] = base_cache_stats.saved_estimated_cost
  return record


def decode_base_cache_stats(record) -> stat_types.BaseCacheStats:
  base_cache_stats = stat_types.BaseCacheStats()
  if 'total_cache_hit' in record:
    base_cache_stats.total_cache_hit = record['total_cache_hit']
  if 'total_success_return' in record:
    base_cache_stats.total_success_return = record['total_success_return']
  if 'total_fail_return' in record:
    base_cache_stats.total_fail_return = record['total_fail_return']
  if 'saved_token_count' in record:
    base_cache_stats.saved_token_count = record['saved_token_count']
  if 'saved_query_token_count' in record:
    base_cache_stats.saved_query_token_count = record['saved_query_token_count']
  if 'saved_response_token_count' in record:
    base_cache_stats.saved_response_token_count = (
        record['saved_response_token_count'])
  if 'saved_total_response_time' in record:
    base_cache_stats.saved_total_response_time = (
        record['saved_total_response_time'])
  if 'saved_estimated_cost' in record:
    base_cache_stats.saved_estimated_cost = record['saved_estimated_cost']
  return base_cache_stats


def encode_provider_model_stats(
    provider_model_stats: stat_types.ProviderModelStats) -> Dict[str, Any]:
  record = {}
  if provider_model_stats.provider_model:
    record['provider_model'] = encode_provider_model_type(
        provider_model_stats.provider_model)
  if provider_model_stats.provider_stats:
    record['provider_stats'] = encode_base_provider_stats(
        provider_model_stats.provider_stats)
  if provider_model_stats.cache_stats:
    record['cache_stats'] = encode_base_cache_stats(
        provider_model_stats.cache_stats)
  return record


def decode_provider_model_stats(
    record: Dict[str, Any]) -> stat_types.ProviderModelStats:
  provider_model_stats = stat_types.ProviderModelStats()
  if 'provider_model' in record:
    provider_model_stats.provider_model = decode_provider_model_type(
        record['provider_model'])
  if 'provider_stats' in record:
    provider_model_stats.provider_stats = decode_base_provider_stats(
        record['provider_stats'])
  if 'cache_stats' in record:
    provider_model_stats.cache_stats = decode_base_cache_stats(
        record['cache_stats'])
  return provider_model_stats


def encode_provider_stats(
    provider_stats: stat_types.ProviderStats) -> Dict[str, Any]:
  record = {}
  if provider_stats.provider:
    record['provider'] = provider_stats.provider
  if provider_stats.provider_stats:
    record['provider_stats'] = encode_base_provider_stats(
        provider_stats.provider_stats)
  if provider_stats.cache_stats:
    record['cache_stats'] = encode_base_cache_stats(provider_stats.cache_stats)
  if provider_stats.provider_models:
    record['provider_models'] = []
    for k, v in provider_stats.provider_models.items():
      value = encode_provider_model_type(k)
      value['provider_model_stats'] = encode_provider_model_stats(v)
      record['provider_models'].append(value)
  return record


def decode_provider_stats(record: Dict[str, Any]) -> stat_types.ProviderStats:
  provider_stats = stat_types.ProviderStats()
  if 'provider' in record:
    provider_stats.provider = record['provider']
  if 'provider_stats' in record:
    provider_stats.provider_stats = decode_base_provider_stats(
        record['provider_stats'])
  if 'cache_stats' in record:
    provider_stats.cache_stats = decode_base_cache_stats(
        record['cache_stats'])
  if 'provider_models' in record:
    provider_stats.provider_models = {}
    for provider_model_record in record['provider_models']:
      provider_model_type = decode_provider_model_type({
          'provider': provider_model_record['provider'],
          'model': provider_model_record['model'],
          'provider_model_identifier': provider_model_record[
              'provider_model_identifier']
      })
      provider_stats.provider_models[provider_model_type] = (
          decode_provider_model_stats(
              provider_model_record['provider_model_stats']))
  return provider_stats


def encode_run_stats(
    run_stats: stat_types.RunStats) -> Dict[str, Any]:
  record = {}
  if run_stats.provider_stats:
    record['provider_stats'] = encode_base_provider_stats(
        run_stats.provider_stats)
  if run_stats.cache_stats:
    record['cache_stats'] = encode_base_cache_stats(run_stats.cache_stats)
  if run_stats.providers:
    record['providers'] = {}
    for k, v in run_stats.providers.items():
      record['providers'][k] = encode_provider_stats(v)
  return record


def decode_run_stats(
    record: Dict[str, Any]) -> stat_types.RunStats:
  run_stats = stat_types.RunStats()
  if 'provider_stats' in record:
    run_stats.provider_stats = decode_base_provider_stats(
        record['provider_stats'])
  if 'cache_stats' in record:
    run_stats.cache_stats = decode_base_cache_stats(record['cache_stats'])
  if 'providers' in record:
    run_stats.providers = {}
    for k, v in record['providers'].items():
      run_stats.providers[k] = decode_provider_stats(v)
  return run_stats



from typing import Optional, List, Dict


class HuggingFaceMock(object):
  def generate_content(
      self,
      messages: List[Dict[str, str]],
      model: str,
      max_tokens: Optional[int]=None,
      temperature: Optional[float]=None,
      stop: Optional[List[str]]=None) -> str:
    return 'mock response'



import copy
import functools
import os
from google import genai
from google.genai import types as genai_types
import proxai.types as types
import proxai.connectors.providers.gemini_mock as gemini_mock
import proxai.connectors.model_connector as model_connector
import proxai.connectors.model_configs as model_configs


class GeminiConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'gemini'

  def init_model(self):
    return genai.Client(
      api_key=os.environ['GEMINI_API_KEY']
    )

  def init_mock_model(self):
    return gemini_mock.GeminiMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: Gemini uses 'user' and 'model' as roles.  'system_instruction' is a
    # different parameter.
    contents = []
    if query_record.prompt is not None:
      contents.append(genai_types.Content(
          parts=[genai_types.Part(text=query_record.prompt)], role='user'))
    if query_record.messages is not None:
      for message in query_record.messages:
        if message['role'] == 'assistant':
          contents.append(genai_types.Content(
              parts=[genai_types.Part(text=message['content'])],
              role='model'))
        if message['role'] == 'user':
          contents.append(genai_types.Content(
              parts=[genai_types.Part(text=message['content'])],
              role='user'))

    config = genai_types.GenerateContentConfig()
    if query_record.system is not None:
      config.system_instruction = query_record.system
    if query_record.max_tokens is not None:
      config.max_output_tokens = query_record.max_tokens
    if query_record.temperature is not None:
      config.temperature = query_record.temperature
    if query_record.stop is not None:
      if isinstance(query_record.stop, str):
        config.stop_sequences = [query_record.stop]
      else:
        config.stop_sequences = query_record.stop
    response = self.api.models.generate_content(
        model=self.provider_model.provider_model_identifier,
        config=config,
        contents=contents
    )
    return response.text



import os
import copy
import datetime
from typing import Optional, Callable
import proxai.types as types
import proxai.serializers.type_serializer as type_serializer
import json
import proxai.state_controllers.state_controller as state_controller

AVAILABLE_MODELS_PATH = 'available_models.json'
_MODEL_CACHE_MANAGER_STATE_PROPERTY = '_model_cache_manager_state'


class ModelCacheManager(state_controller.StateControlled):
  _cache_options: types.CacheOptions
  _model_status_by_call_type: types.ModelStatusByCallType
  _model_cache_manager_state: types.ModelCacheManagerState

  def __init__(
      self,
      cache_options: Optional[types.CacheOptions] = None,
      get_cache_options: Optional[Callable[[], types.CacheOptions]] = None,
      init_state: Optional[types.ModelCacheManagerState] = None):

    if init_state and (
        cache_options is not None or
        get_cache_options is not None):
      raise ValueError(
          'init_state and other parameters cannot be set at the same time.')

    super().__init__(
        cache_options=cache_options,
        get_cache_options=get_cache_options)

    self.init_state()
    self.set_property_value(
        'status', types.ModelCacheManagerStatus.INITIALIZING)

    if init_state:
      self.load_state(init_state)
    else:
      initial_state = self.get_state()
      self._get_cache_options = get_cache_options
      self.cache_options = cache_options
      self.handle_changes(initial_state, self.get_state())

  def get_internal_state_property_name(self):
    return _MODEL_CACHE_MANAGER_STATE_PROPERTY

  def get_internal_state_type(self):
    return types.ModelCacheManagerState

  def handle_changes(
      self,
      old_state: types.ModelCacheManagerState,
      current_state: types.ModelCacheManagerState):

    if current_state.cache_options is None:
      self.status = types.ModelCacheManagerStatus.CACHE_OPTIONS_NOT_FOUND
      self.model_status_by_call_type = None
      return

    if current_state.cache_options.disable_model_cache == True:
      self.status = types.ModelCacheManagerStatus.DISABLED
      self.model_status_by_call_type = None
      return

    if current_state.cache_options.cache_path is None:
      self.status = types.ModelCacheManagerStatus.CACHE_PATH_NOT_FOUND
      self.model_status_by_call_type = None
      return

    if not os.access(current_state.cache_options.cache_path, os.W_OK):
      self.status = types.ModelCacheManagerStatus.CACHE_PATH_NOT_WRITABLE
      self.model_status_by_call_type = None
      return

    if current_state == old_state:
      return

    if (old_state.cache_options is None or
        old_state.cache_options.cache_path !=
        current_state.cache_options.cache_path or
        old_state.cache_options.disable_model_cache !=
        current_state.cache_options.disable_model_cache or
        old_state.cache_options.clear_model_cache_on_connect !=
        current_state.cache_options.clear_model_cache_on_connect or
        old_state.cache_options.model_cache_duration !=
        current_state.cache_options.model_cache_duration):
      self._load_from_cache_path()

    self.status = types.ModelCacheManagerStatus.WORKING

  @property
  def cache_path(self) -> str:
    if self.cache_options is None or self.cache_options.cache_path is None:
      return None
    return os.path.join(self.cache_options.cache_path, AVAILABLE_MODELS_PATH)

  @cache_path.setter
  def cache_path(self, value: str):
    raise ValueError('cache_path needs to be set through cache_options.')

  @property
  def status(self) -> types.ModelCacheManagerStatus:
    return self.get_property_value('status')

  @status.setter
  def status(self, value: types.ModelCacheManagerStatus):
    self.set_property_value('status', value)

  @property
  def cache_options(self) -> types.CacheOptions:
    return self.get_property_value('cache_options')

  @cache_options.setter
  def cache_options(self, value: types.CacheOptions):
    self.set_property_value('cache_options', value)

  @property
  def model_status_by_call_type(self) -> types.ModelStatusByCallType:
    if getattr(self, '_model_status_by_call_type', None) is None:
      self._model_status_by_call_type = {}
    return self._model_status_by_call_type

  @model_status_by_call_type.setter
  def model_status_by_call_type(self, value: types.ModelStatusByCallType):
    self._model_status_by_call_type = value

  def clear_cache(self):
    self.model_status_by_call_type = None
    if self.cache_path is None:
      return
    if os.path.exists(self.cache_path):
      os.remove(self.cache_path)

  def _save_to_cache_path(self):
    if self.cache_path is None:
      return
    data = copy.deepcopy(self.model_status_by_call_type)
    for call_value in data.keys():
      data[call_value] = type_serializer.encode_model_status(data[call_value])
    with open(self.cache_path, 'w') as f:
      json.dump(data, f)

  def _load_from_cache_path(self):
    self.model_status_by_call_type = {}
    if self.cache_path is None:
      return
    if not os.path.exists(self.cache_path):
      return
    data = {}
    with open(self.cache_path, 'r') as f:
      try:
        data = json.load(f)
        for call_value in data.keys():
          self.model_status_by_call_type[
              call_value] = type_serializer.decode_model_status(data[call_value])
      except Exception as e:
        error_message = (
            '_load_from_cache_path failed because of the parsing error.\n'
            '* Please check cache path is correct.\n'
            f'    > Cache path: {self.cache_path}\n'
            '* Try to clean the model cache on px.connect:\n'
            '    > px.connect(\n'
            '    >    cache_options=px.CacheOptions(\n'
            '    >        clear_model_cache_on_connect=True))\n'
            '* If the problem persists, delete the cache file and try again:\n'
            f'    > rm {self.cache_path};\n'
            '* Open bug report at https://github.com/proxai/proxai/issues')
        raise ValueError(error_message)

  def _clean_model_from_tested_models(
      self,
      call_type: types.CallType,
      model: types.ProviderModelType):
    """Cleans a model from the tested models: working_models and failed_models.
    """
    if call_type not in self.model_status_by_call_type:
      return
    model_status = self.model_status_by_call_type[call_type]
    if model in model_status.working_models:
      model_status.working_models.remove(model)
    if model in model_status.failed_models:
      model_status.failed_models.remove(model)
    model_status.provider_queries.pop(model, None)

  def _clean_model_from_model_status(
      self,
      call_type: types.CallType,
      model: types.ProviderModelType):
    if call_type not in self.model_status_by_call_type:
      return
    model_status = self.model_status_by_call_type[call_type]
    model_status.unprocessed_models.discard(model)
    model_status.working_models.discard(model)
    model_status.failed_models.discard(model)
    model_status.filtered_models.discard(model)
    model_status.provider_queries.pop(model, None)

  def get(self, call_type: types.CallType) -> types.ModelStatus:
    result = types.ModelStatus()
    if call_type not in self.model_status_by_call_type:
      return result

    if self.cache_options.model_cache_duration is None:
      return self.model_status_by_call_type[call_type]

    model_status = self.model_status_by_call_type[call_type]
    for model in list(model_status.provider_queries.keys()):
      if model not in model_status.provider_queries:
        continue
      provider_query = model_status.provider_queries[model]
      time_since_response = (datetime.datetime.now(datetime.timezone.utc)
                             - provider_query.response_record.end_utc_date
                             ).total_seconds()
      if time_since_response > self.cache_options.model_cache_duration:
        self._clean_model_from_tested_models(call_type=call_type, model=model)

    self._save_to_cache_path()
    return model_status

  def update(
      self,
      model_status_updates: types.ModelStatus,
      call_type: types.CallType):
    if call_type not in self.model_status_by_call_type:
      self.model_status_by_call_type[call_type] = types.ModelStatus()
    all_updated_models = (
        model_status_updates.unprocessed_models |
        model_status_updates.working_models |
        model_status_updates.failed_models |
        model_status_updates.filtered_models)
    for provider_query_model in model_status_updates.provider_queries.keys():
      if provider_query_model not in all_updated_models:
        raise ValueError(
            f'Model {provider_query_model} is not in any of the unprocessed, '
            'working, failed, or filtered models. Please provide the provider '
            'model in one of the sets when updating provider_queries.')

    for model in all_updated_models:
      self._clean_model_from_model_status(call_type=call_type, model=model)

    model_status = self.model_status_by_call_type[call_type]
    for model in model_status_updates.unprocessed_models:
      model_status.unprocessed_models.add(model)
    for model in model_status_updates.working_models:
      model_status.working_models.add(model)
    for model in model_status_updates.failed_models:
      model_status.failed_models.add(model)
    for model in model_status_updates.filtered_models:
      model_status.filtered_models.add(model)
    for provider_model, provider_query in (
        model_status_updates.provider_queries.items()):
      model_status.provider_queries[provider_model] = provider_query
    self._save_to_cache_path()

  def save(
      self,
      model_status: types.ModelStatus,
      call_type: types.CallType):
    self.model_status_by_call_type[call_type] = copy.deepcopy(model_status)
    self._save_to_cache_path()



import os
import copy
from datetime import datetime
from typing import Dict, Optional
import proxai.types as types
import proxai.serializers.type_serializer as type_serializer
import json
from pprint import pprint

QUERY_LOGGING_FILE_NAME = 'provider_queries.log'
ERROR_LOGGING_FILE_NAME = 'errors.log'
WARNING_LOGGING_FILE_NAME = 'warnings.log'
INFO_LOGGING_FILE_NAME = 'info.log'
MERGED_LOGGING_FILE_NAME = 'merged.log'
PROXDASH_LOGGING_FILE_NAME = 'proxdash.log'
_SENSITIVE_CONTENT_HIDDEN_STRING = '<sensitive content hidden>'


def _hide_sensitive_content_query_record(
    query_record: types.QueryRecord) -> types.QueryRecord:
  query_record = copy.deepcopy(query_record)
  if query_record.system:
    query_record.system = _SENSITIVE_CONTENT_HIDDEN_STRING
  if query_record.prompt:
    query_record.prompt = _SENSITIVE_CONTENT_HIDDEN_STRING
  if query_record.messages:
    query_record.messages = [
      {
        'role': 'assistant',
        'content': _SENSITIVE_CONTENT_HIDDEN_STRING
      }
    ]
  return query_record


def _hide_sensitive_content_query_response_record(
    query_response_record: types.QueryResponseRecord) -> types.QueryResponseRecord:
  query_response_record = copy.deepcopy(query_response_record)
  if query_response_record.response:
    query_response_record.response = _SENSITIVE_CONTENT_HIDDEN_STRING
  return query_response_record


def _hide_sensitive_content_logging_record(
    logging_record: types.LoggingRecord) -> types.LoggingRecord:
  logging_record = copy.deepcopy(logging_record)
  if logging_record.query_record:
    logging_record.query_record = _hide_sensitive_content_query_record(
        logging_record.query_record)
  if logging_record.response_record:
    logging_record.response_record = (
        _hide_sensitive_content_query_response_record(
            logging_record.response_record))
  return logging_record


def _write_log(
    logging_options: types.LoggingOptions,
    file_name: str,
    data: Dict):
    file_path = os.path.join(logging_options.logging_path, file_name)
    with open(file_path, 'a') as f:
      f.write(json.dumps(data) + '\n')
    f.close()


def log_logging_record(
    logging_options: types.LoggingOptions,
    logging_record: types.LoggingRecord):
  if not logging_options:
    return
  if logging_options.hide_sensitive_content:
    logging_record = _hide_sensitive_content_logging_record(logging_record)
  result = type_serializer.encode_logging_record(logging_record)
  if logging_options.stdout:
    pprint(result)
  if not logging_options.logging_path:
    return
  _write_log(
      logging_options=logging_options,
      file_name=QUERY_LOGGING_FILE_NAME,
      data=result)


def log_message(
    logging_options: types.LoggingOptions,
    message: str,
    type: types.LoggingType,
    query_record: Optional[types.QueryRecord] = None):
  if not logging_options:
    return
  result = {}
  result['logging_type'] = type.value.upper()
  result['message'] = message
  result['timestamp'] = datetime.now().isoformat()
  if query_record:
    if logging_options.hide_sensitive_content:
      query_record = _hide_sensitive_content_query_record(query_record)
    result['query_record'] = type_serializer.encode_query_record(query_record)
  if logging_options.stdout:
    pprint(result)
  if not logging_options.logging_path:
    return
  if type == types.LoggingType.ERROR:
    _write_log(
        logging_options=logging_options,
        file_name=ERROR_LOGGING_FILE_NAME,
        data=result)
  elif type == types.LoggingType.WARNING:
    _write_log(
        logging_options=logging_options,
        file_name=WARNING_LOGGING_FILE_NAME,
        data=result)
  else:
    _write_log(
        logging_options=logging_options,
        file_name=INFO_LOGGING_FILE_NAME,
        data=result)
  _write_log(
      logging_options=logging_options,
      file_name=MERGED_LOGGING_FILE_NAME,
      data=result)


def log_proxdash_message(
    logging_options: types.LoggingOptions,
    proxdash_options: types.ProxDashOptions,
    message: str,
    type: types.LoggingType,
    query_record: Optional[types.QueryRecord] = None):
  result = {}
  result['logging_type'] = type.value.upper()
  result['message'] = message
  result['timestamp'] = datetime.now().isoformat()
  if query_record:
    if logging_options.hide_sensitive_content:
      query_record = _hide_sensitive_content_query_record(query_record)
    result['query_record'] = type_serializer.encode_query_record(query_record)
  if proxdash_options.stdout:
    pprint(result)
  if not logging_options.logging_path:
    return
  if type == types.LoggingType.ERROR:
    _write_log(
        logging_options=logging_options,
        file_name=ERROR_LOGGING_FILE_NAME,
        data=result)
  elif type == types.LoggingType.WARNING:
    _write_log(
        logging_options=logging_options,
        file_name=WARNING_LOGGING_FILE_NAME,
        data=result)
  _write_log(
      logging_options=logging_options,
      file_name=PROXDASH_LOGGING_FILE_NAME,
      data=result)
  _write_log(
      logging_options=logging_options,
      file_name=MERGED_LOGGING_FILE_NAME,
      data=result)



from types import MappingProxyType
from typing import Dict
import proxai.types as types

GENERATE_TEXT_FEATURES: Dict[
    str, Dict[str, types.ProviderModelFeatureType]] = MappingProxyType({
  'mistral': MappingProxyType({
    'ministral-3b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'ministral-8b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'open-mistral-7b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'open-mistral-nemo': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'open-mixtral-8x7b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'open-mixtral-8x22b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'mistral-small': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'mistral-large': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'pixtral-large': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'codestral': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'pixtral-12b': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
    'mistral-saba': types.ProviderModelFeatureType(
      not_supported_features=['stop'],
    ),
  }),

  'openai': MappingProxyType({
    'gpt-4o-search-preview': types.ProviderModelFeatureType(
      not_supported_features=['temperature', 'stop'],
    ),
    'gpt-4o-mini-search-preview': types.ProviderModelFeatureType(
      not_supported_features=['temperature', 'stop'],
    ),
    'o1-mini': types.ProviderModelFeatureType(
      not_supported_features=['system', 'max_tokens', 'temperature', 'stop'],
    ),
    'o1': types.ProviderModelFeatureType(
      not_supported_features=['temperature'],
    ),
    'o3-mini': types.ProviderModelFeatureType(
      not_supported_features=['max_tokens', 'temperature'],
    ),
    'o4-mini': types.ProviderModelFeatureType(
      not_supported_features=['max_tokens', 'temperature', 'stop'],
    ),
  }),
})



import copy
import dataclasses
import enum
from typing import Dict, Optional
import proxai.types as types


@dataclasses.dataclass
class BaseProviderStats:
  total_queries: int = 0
  total_successes: int = 0
  total_fails: int = 0

  total_token_count: int = 0
  total_query_token_count: int = 0
  total_response_token_count: int = 0

  _total_response_time: float = 0.0
  _avr_response_time: float = 0.0

  estimated_cost: int = 0.0

  total_cache_look_fail_reasons: Dict[types.CacheLookFailReason, int] = (
      dataclasses.field(default_factory=dict))

  def __init__(
      self,
      total_queries: int = 0,
      total_successes: int = 0,
      total_fails: int = 0,
      total_token_count: int = 0,
      total_query_token_count: int = 0,
      total_response_token_count: int = 0,
      total_response_time: float = 0.0,
      estimated_cost: int = 0.0,
      total_cache_look_fail_reasons: Dict[types.CacheLookFailReason, int] = {},
  ):
    self.total_queries = total_queries
    self.total_successes = total_successes
    self.total_fails = total_fails
    self.total_token_count = total_token_count
    self.total_query_token_count = total_query_token_count
    self.total_response_token_count = total_response_token_count
    self.total_response_time = total_response_time
    self.estimated_cost = estimated_cost
    self.total_cache_look_fail_reasons = total_cache_look_fail_reasons

  @property
  def total_response_time(self):
    return self._total_response_time

  @total_response_time.setter
  def total_response_time(self, value):
    self._total_response_time = value
    if self.total_successes == 0:
      self._avr_response_time = 0.0
    else:
      self._avr_response_time = (
          self.total_response_time / self.total_successes)

  @property
  def avr_response_time(self):
    return self._avr_response_time

  @avr_response_time.setter
  def avr_response_time(self, value):
    raise ValueError('avr_response_time is a calculated property.')

  def __add__(self, other):
    if not isinstance(other, BaseProviderStats):
      raise ValueError(
          'Invalid addition of BaseProviderStats objects.'
          f'Cannot add BaseProviderStats with {type(other)}')
    total_cache_look_fail_reasons = copy.deepcopy(
        self.total_cache_look_fail_reasons)
    for k, v in other.total_cache_look_fail_reasons.items():
      if k in total_cache_look_fail_reasons:
        total_cache_look_fail_reasons[k] += v
      else:
        total_cache_look_fail_reasons[k] = v
    return BaseProviderStats(
        total_queries=self.total_queries + other.total_queries,
        total_successes=self.total_successes + other.total_successes,
        total_fails=self.total_fails + other.total_fails,
        total_token_count=self.total_token_count + other.total_token_count,
        total_query_token_count=(
            self.total_query_token_count + other.total_query_token_count),
        total_response_token_count=(
            self.total_response_token_count + other.total_response_token_count),
        total_response_time=(
            self.total_response_time + other.total_response_time),
        estimated_cost=self.estimated_cost + other.estimated_cost,
        total_cache_look_fail_reasons=total_cache_look_fail_reasons)

  def __sub__(self, other):
    if not isinstance(other, BaseProviderStats):
      raise ValueError(
          'Invalid subtraction of BaseProviderStats objects.'
          f'Cannot subtract BaseProviderStats with {type(other)}')
    total_cache_look_fail_reasons = copy.deepcopy(
        self.total_cache_look_fail_reasons)
    for k, v in other.total_cache_look_fail_reasons.items():
      if k in total_cache_look_fail_reasons:
        total_cache_look_fail_reasons[k] -= v
      else:
        raise ValueError(
            f'Invalid subtraction of BaseCacheStats objects.'
            f'Key {k} not found in total_cache_look_fail_reasons.')
    for k in list(total_cache_look_fail_reasons.keys()):
      if total_cache_look_fail_reasons[k] <= 0:
        del total_cache_look_fail_reasons[k]

    result = BaseProviderStats(
        total_queries=self.total_queries - other.total_queries,
        total_successes=self.total_successes - other.total_successes,
        total_fails=self.total_fails - other.total_fails,
        total_token_count=self.total_token_count - other.total_token_count,
        total_query_token_count=(
            self.total_query_token_count - other.total_query_token_count),
        total_response_token_count=(
            self.total_response_token_count - other.total_response_token_count),
        total_response_time=(
            self.total_response_time - other.total_response_time),
        estimated_cost=self.estimated_cost - other.estimated_cost,
        total_cache_look_fail_reasons=total_cache_look_fail_reasons)
    for key in result.__dict__:
      if isinstance(getattr(result, key), dict):
        for k in getattr(result, key):
          if getattr(result, key)[k] < -0.0001:
            raise ValueError(
                f'Invalid subtraction of BaseCacheStats objects.'
                f'Negative value for {key}[{k}]: {getattr(result, key)[k]}')
          if getattr(result, key)[k] < 0:
            result.__dict__[key][k] = 0
      else:
        if getattr(result, key) < -0.0001:
          raise ValueError(
              f'Invalid subtraction of BaseCacheStats objects.'
              f'Negative value for {key}: {getattr(result, key)}')
        if getattr(result, key) < 0:
          result.__dict__[key] = 0
    return result


@dataclasses.dataclass
class BaseCacheStats:
  total_cache_hit: int = 0
  total_success_return: int = 0
  total_fail_return: int = 0

  saved_token_count: int = 0
  saved_query_token_count: int = 0
  saved_response_token_count: int = 0

  _saved_total_response_time: float = 0.0
  _saved_avr_response_time: float = 0.0

  saved_estimated_cost: int = 0.0

  def __init__(
      self,
      total_cache_hit: int = 0,
      total_success_return: int = 0,
      total_fail_return: int = 0,
      saved_token_count: int = 0,
      saved_query_token_count: int = 0,
      saved_response_token_count: int = 0,
      saved_total_response_time: float = 0.0,
      saved_estimated_cost: int = 0.0,
  ):
    self.total_cache_hit = total_cache_hit
    self.total_success_return = total_success_return
    self.total_fail_return = total_fail_return
    self.saved_token_count = saved_token_count
    self.saved_query_token_count = saved_query_token_count
    self.saved_response_token_count = saved_response_token_count
    self.saved_total_response_time = saved_total_response_time
    self.saved_estimated_cost = saved_estimated_cost

  @property
  def saved_total_response_time(self):
    return self._saved_total_response_time

  @saved_total_response_time.setter
  def saved_total_response_time(self, value):
    self._saved_total_response_time = value
    if self.total_success_return == 0:
      self._saved_avr_response_time = 0.0
    else:
      self._saved_avr_response_time = (
          self.saved_total_response_time / self.total_success_return)

  @property
  def saved_avr_response_time(self):
    return self._saved_avr_response_time

  @saved_avr_response_time.setter
  def saved_avr_response_time(self, value):
    raise ValueError('saved_avr_response_time is a calculated property.')

  def __add__(self, other):
    if not isinstance(other, BaseCacheStats):
      raise ValueError(
          'Invalid addition of BaseCacheStats objects.'
          f'Cannot add BaseCacheStats with {type(other)}')
    return BaseCacheStats(
        total_cache_hit=self.total_cache_hit + other.total_cache_hit,
        total_success_return=(
            self.total_success_return + other.total_success_return),
        total_fail_return=self.total_fail_return + other.total_fail_return,
        saved_token_count=self.saved_token_count + other.saved_token_count,
        saved_query_token_count=(
            self.saved_query_token_count + other.saved_query_token_count),
        saved_response_token_count=(
            self.saved_response_token_count + other.saved_response_token_count),
        saved_total_response_time=(
            self.saved_total_response_time + other.saved_total_response_time),
        saved_estimated_cost=(
            self.saved_estimated_cost + other.saved_estimated_cost))

  def __sub__(self, other):
    if not isinstance(other, BaseCacheStats):
      raise ValueError(
          'Invalid subtraction of BaseCacheStats objects.'
          f'Cannot subtract BaseCacheStats with {type(other)}')

    result = BaseCacheStats(
        total_cache_hit=self.total_cache_hit - other.total_cache_hit,
        total_success_return=(
            self.total_success_return - other.total_success_return),
        total_fail_return=self.total_fail_return - other.total_fail_return,
        saved_token_count=self.saved_token_count - other.saved_token_count,
        saved_query_token_count=(
            self.saved_query_token_count - other.saved_query_token_count),
        saved_response_token_count=(
            self.saved_response_token_count - other.saved_response_token_count),
        saved_total_response_time=(
            self.saved_total_response_time - other.saved_total_response_time),
        saved_estimated_cost=(
            self.saved_estimated_cost - other.saved_estimated_cost))

    for key in result.__dict__:
      if getattr(result, key) < -0.0001:
        raise ValueError(
            f'Invalid subtraction of BaseProviderStats objects.'
            f'Negative value for {key}: {getattr(result, key)}')
      if getattr(result, key) < 0:
        result.__dict__[key] = 0
    return result


@dataclasses.dataclass
class BaseStats:
  provider_stats: Optional[BaseProviderStats] = None
  cache_stats: Optional[BaseCacheStats] = None

  def _add_base_provider_stats(self, other: BaseProviderStats):
    if self.provider_stats is None:
      self.provider_stats = BaseProviderStats()
    if other:
      self.provider_stats += other

  def _sub_base_provider_stats(self, other: BaseProviderStats):
    if self.provider_stats is None:
      self.provider_stats = BaseProviderStats()
    if other:
      self.provider_stats -= other

  def _add_base_cache_stats(self, other: BaseCacheStats):
    if self.cache_stats is None:
      self.cache_stats = BaseCacheStats()
    if other:
      self.cache_stats += other

  def _sub_base_cache_stats(self, other: BaseCacheStats):
    if self.cache_stats is None:
      self.cache_stats = BaseCacheStats()
    if other:
      self.cache_stats -= other


@dataclasses.dataclass
class ProviderModelStats(BaseStats):
  provider_model: Optional[types.ProviderModelType] = None

  def __add__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, BaseProviderStats):
      result._add_base_provider_stats(other)
    elif isinstance(other, BaseCacheStats):
      result._add_base_cache_stats(other)
    elif isinstance(other, ProviderModelStats):
      if self.provider_model != other.provider_model:
        raise ValueError(
            'Cannot add ProviderModelStats with different provider_model types.'
            f'{self.provider_model} != {other.provider_model}')
      result._add_base_provider_stats(other.provider_stats)
      result._add_base_cache_stats(other.cache_stats)
    else:
      raise ValueError(
          'Invalid addition of ProviderModelStats objects.'
          f'Cannot add ProviderModelStats with {type(other)}')
    return result

  def __sub__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, BaseProviderStats):
      result._sub_base_provider_stats(other)
    elif isinstance(other, BaseCacheStats):
      result._sub_base_cache_stats(other)
    elif isinstance(other, ProviderModelStats):
      if self.provider_model != other.provider_model:
        raise ValueError(
            'Cannot subtract ProviderModelStats with different provider_model '
            f'types. {self.provider_model} != {other.provider_model}')
      result._sub_base_provider_stats(other.provider_stats)
      result._sub_base_cache_stats(other.cache_stats)
    else:
      raise ValueError(
          'Invalid subtraction of ProviderModelStats objects.'
          f'Cannot subtract ProviderModelStats with {type(other)}')
    return result


@dataclasses.dataclass
class ProviderStats(BaseStats):
  provider: Optional[str] = None
  provider_models: Optional[
      Dict[types.ProviderModelType, ProviderModelStats]] = None

  def _add_provider_model_stats(self, other: ProviderModelStats):
    self._add_base_provider_stats(other.provider_stats)
    self._add_base_cache_stats(other.cache_stats)
    if self.provider_models is None:
      self.provider_models = {}
    if other.provider_model and other.provider_model not in self.provider_models:
      self.provider_models[other.provider_model] = ProviderModelStats(
          provider_model=other.provider_model)
    self.provider_models[other.provider_model] += other
    return self.provider_models

  def _sub_provider_model_stats(self, other: ProviderModelStats):
    self._sub_base_provider_stats(other.provider_stats)
    self._sub_base_cache_stats(other.cache_stats)
    if self.provider_models is None:
      self.provider_models = {}
    if other.provider_model and other.provider_model not in self.provider_models:
      self.provider_models[other.provider_model] = ProviderModelStats(
          provider_model=other.provider_model)
    self.provider_models[other.provider_model] -= other
    return self.provider_models

  def __add__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, ProviderModelStats):
      result._add_provider_model_stats(other)
    elif isinstance(other, ProviderStats):
      if self.provider != other.provider:
        raise ValueError(
            'Cannot add ProviderStats with different provider types.'
            f'{self.provider} != {other.provider}')
      for model_stat in other.provider_models.values():
        result._add_provider_model_stats(model_stat)
    else:
      raise ValueError(
          'Invalid addition of ProviderStats objects.'
          f'Cannot add ProviderStats with {type(other)}')
    return result

  def __sub__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, ProviderModelStats):
      result._sub_provider_model_stats(other)
    elif isinstance(other, ProviderStats):
      if self.provider != other.provider:
        raise ValueError(
            'Cannot subtract ProviderStats with different provider types.'
            f'{self.provider} != {other.provider}')
      for model_stat in other.provider_models.values():
        result._sub_provider_model_stats(model_stat)
      for model in list(result.provider_models.keys()):
        if (result.provider_models[model].provider_stats.total_queries == 0 and
            result.provider_models[model].cache_stats.total_cache_hit == 0):
          del result.provider_models[model]
    else:
      raise ValueError(
          'Invalid subtraction of ProviderStats objects.'
          f'Cannot subtract ProviderStats with {type(other)}')
    return result


class GlobalStatType(str, enum.Enum):
  RUN_TIME = 'run_time'
  SINCE_CONNECT = 'since_connect'


@dataclasses.dataclass
class RunStats(BaseStats):
  providers: Dict[str, ProviderStats] = dataclasses.field(
      default_factory=dict)

  def _add_provider_model_stats(self, other: ProviderModelStats):
    self._add_base_provider_stats(other.provider_stats)
    self._add_base_cache_stats(other.cache_stats)
    if self.providers is None:
      self.providers = {}
    if other.provider_model.provider not in self.providers:
      self.providers[other.provider_model.provider] = ProviderStats(
          provider=other.provider_model.provider)
    self.providers[other.provider_model.provider] += other

  def _sub_provider_model_stats(self, other: ProviderModelStats):
    self._sub_base_provider_stats(other.provider_stats)
    self._sub_base_cache_stats(other.cache_stats)
    if self.providers is None:
      self.providers = {}
    if other.provider_model.provider not in self.providers:
      self.providers[other.provider_model.provider] = ProviderStats(
          provider=other.provider_model.provider)
    self.providers[other.provider_model.provider] -= other
    for provider in list(self.providers.keys()):
      if (self.providers[provider].provider_stats.total_queries == 0 and
          self.providers[provider].cache_stats.total_cache_hit == 0):
        del self.providers[provider]

  def _add_provider_stats(self, other: ProviderStats):
    self._add_base_provider_stats(other.provider_stats)
    self._add_base_cache_stats(other.cache_stats)
    if self.providers is None:
      self.providers = {}
    if other.provider not in self.providers:
      self.providers[other.provider] = ProviderStats(provider=other.provider)
    self.providers[other.provider] += other

  def _sub_provider_stats(self, other: ProviderStats):
    self._sub_base_provider_stats(other.provider_stats)
    self._sub_base_cache_stats(other.cache_stats)
    if self.providers is None:
      self.providers = {}
    if other.provider not in self.providers:
      self.providers[other.provider] = ProviderStats(provider=other.provider)
    self.providers[other.provider] -= other
    for provider in list(self.providers.keys()):
      if (self.providers[provider].provider_stats.total_queries == 0 and
          self.providers[provider].cache_stats.total_cache_hit == 0):
        del self.providers[provider]

  def __add__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, ProviderModelStats):
      result._add_provider_model_stats(other)
    elif isinstance(other, ProviderStats):
      result._add_provider_stats(other)
    elif isinstance(other, RunStats):
      for provider_stat in other.providers.values():
        result._add_provider_stats(provider_stat)
    else:
      raise ValueError(
          'Invalid addition of RunStats objects.'
          f'Cannot add RunStats with {type(other)}')
    return result

  def __sub__(self, other):
    result = copy.deepcopy(self)
    if isinstance(other, ProviderModelStats):
      result._sub_provider_model_stats(other)
    elif isinstance(other, ProviderStats):
      result._sub_provider_stats(other)
    elif isinstance(other, RunStats):
      for provider_stat in other.providers.values():
        result._sub_provider_stats(provider_stat)
      for provider in list(result.providers.keys()):
        if (result.providers[provider].provider_stats.total_queries == 0 and
            result.providers[provider].cache_stats.total_cache_hit == 0):
          del result.providers[provider]
    else:
      raise ValueError(
          'Invalid subtraction of RunStats objects.'
          f'Cannot subtract RunStats with {type(other)}')
    return result



import copy
import functools
import os
from openai import OpenAI
import proxai.types as types
import proxai.connectors.providers.openai_mock as openai_mock
import proxai.connectors.model_connector as model_connector


class DeepSeekConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'deepseek'

  def init_model(self):
    return OpenAI(
        api_key=os.getenv("DEEPSEEK_API_KEY"),
        base_url="https://api.deepseek.com")

  def init_mock_model(self):
    return openai_mock.OpenAIMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: OpenAI uses 'system', 'user', and 'assistant' as roles.
    query_messages = []
    if query_record.system is not None:
      query_messages.append({'role': 'system', 'content': query_record.system})
    if query_record.prompt is not None:
      query_messages.append({'role': 'user', 'content': query_record.prompt})
    if query_record.messages is not None:
      query_messages.extend(query_record.messages)
    provider_model = query_record.provider_model

    create = functools.partial(
        self.api.chat.completions.create,
        model=provider_model.provider_model_identifier,
        messages=query_messages)
    if query_record.max_tokens is not None:
      create = functools.partial(
          create, max_completion_tokens=query_record.max_tokens)
    if query_record.temperature is not None:
      create = functools.partial(create, temperature=query_record.temperature)
    if query_record.stop is not None:
      create = functools.partial(create, stop=query_record.stop)

    completion = create()
    return completion.choices[0].message.content



import copy
import datetime
import functools
import os
import tempfile
from typing import Any, Dict, Optional, Tuple, Union
import platformdirs
import proxai.types as types
import proxai.type_utils as type_utils
import proxai.connectors.model_connector as model_connector
import proxai.connectors.model_registry as model_registry
import proxai.caching.query_cache as query_cache
import proxai.caching.model_cache as model_cache
import proxai.serializers.type_serializer as type_serializer
import proxai.stat_types as stat_types
import proxai.connections.available_models as available_models
import proxai.connections.proxdash as proxdash
import proxai.experiment.experiment as experiment
import proxai.connectors.model_configs as model_configs
import proxai.logging.utils as logging_utils

_RUN_TYPE: types.RunType
_HIDDEN_RUN_KEY: str
_EXPERIMENT_PATH: Optional[str]
_ROOT_LOGGING_PATH: Optional[str]
_DEFAULT_MODEL_CACHE_PATH: Optional[tempfile.TemporaryDirectory]
_PLATFORM_USED_FOR_DEFAULT_MODEL_CACHE: bool

_LOGGING_OPTIONS: types.LoggingOptions
_CACHE_OPTIONS: types.CacheOptions
_PROXDASH_OPTIONS: types.ProxDashOptions

_REGISTERED_MODEL_CONNECTORS: Dict[
    types.CallType, model_connector.ProviderModelConnector]
_MODEL_CONNECTORS: Dict[
    types.ProviderModelType, model_connector.ProviderModelConnector]
_DEFAULT_MODEL_CACHE_MANAGER: Optional[model_cache.ModelCacheManager]
_MODEL_CACHE_MANAGER: Optional[model_cache.ModelCacheManager]
_QUERY_CACHE_MANAGER: Optional[query_cache.QueryCacheManager]
_PROXDASH_CONNECTION: Optional[proxdash.ProxDashConnection]

_STRICT_FEATURE_TEST: bool
_SUPPRESS_PROVIDER_ERRORS: bool
_ALLOW_MULTIPROCESSING: bool
_MODEL_TEST_TIMEOUT: Optional[int]

_STATS: Dict[stat_types.GlobalStatType, stat_types.RunStats]
_AVAILABLE_MODELS: Optional[available_models.AvailableModels]

CacheOptions = types.CacheOptions
LoggingOptions = types.LoggingOptions
ProxDashOptions = types.ProxDashOptions


def _init_default_model_cache_manager():
  global _DEFAULT_MODEL_CACHE_PATH
  global _PLATFORM_USED_FOR_DEFAULT_MODEL_CACHE
  global _DEFAULT_MODEL_CACHE_MANAGER
  try:
    app_dirs = platformdirs.PlatformDirs(appname="proxai", appauthor="proxai")
    _DEFAULT_MODEL_CACHE_PATH =  app_dirs.user_cache_dir
    os.makedirs(_DEFAULT_MODEL_CACHE_PATH, exist_ok=True)
    # 4 hours cache duration makes sense for local development if proxai is
    # using platform app cache directory
    _DEFAULT_MODEL_CACHE_MANAGER = model_cache.ModelCacheManager(
        cache_options=types.CacheOptions(
            cache_path=_DEFAULT_MODEL_CACHE_PATH,
            model_cache_duration=60 * 60 * 4))
    _PLATFORM_USED_FOR_DEFAULT_MODEL_CACHE = True
  except Exception as e:
    _DEFAULT_MODEL_CACHE_PATH = tempfile.TemporaryDirectory()
    _DEFAULT_MODEL_CACHE_MANAGER = model_cache.ModelCacheManager(
        cache_options=types.CacheOptions(
            cache_path=_DEFAULT_MODEL_CACHE_PATH.name))
    _PLATFORM_USED_FOR_DEFAULT_MODEL_CACHE = False


def _init_globals():
  global _RUN_TYPE
  global _HIDDEN_RUN_KEY
  global _EXPERIMENT_PATH
  global _ROOT_LOGGING_PATH

  global _LOGGING_OPTIONS
  global _CACHE_OPTIONS
  global _PROXDASH_OPTIONS

  global _REGISTERED_MODEL_CONNECTORS
  global _MODEL_CONNECTORS
  global _MODEL_CACHE_MANAGER
  global _QUERY_CACHE_MANAGER
  global _PROXDASH_CONNECTION

  global _STRICT_FEATURE_TEST
  global _SUPPRESS_PROVIDER_ERRORS
  global _ALLOW_MULTIPROCESSING
  global _MODEL_TEST_TIMEOUT

  global _STATS
  global _AVAILABLE_MODELS

  _RUN_TYPE = types.RunType.PRODUCTION
  _HIDDEN_RUN_KEY = experiment.get_hidden_run_key()
  _EXPERIMENT_PATH = None
  _ROOT_LOGGING_PATH = None

  _LOGGING_OPTIONS = types.LoggingOptions()
  _CACHE_OPTIONS = types.CacheOptions()
  _PROXDASH_OPTIONS = types.ProxDashOptions()

  _REGISTERED_MODEL_CONNECTORS = {}
  _MODEL_CONNECTORS = {}
  _MODEL_CACHE_MANAGER = None
  _QUERY_CACHE_MANAGER = None
  _PROXDASH_CONNECTION = None
  _init_default_model_cache_manager()

  _STRICT_FEATURE_TEST = False
  _SUPPRESS_PROVIDER_ERRORS = False
  _ALLOW_MULTIPROCESSING = True
  _MODEL_TEST_TIMEOUT = 25

  _STATS = {
      stat_types.GlobalStatType.RUN_TIME: stat_types.RunStats(),
      stat_types.GlobalStatType.SINCE_CONNECT: stat_types.RunStats()
  }
  _AVAILABLE_MODELS = None


def _set_experiment_path(
    experiment_path: Optional[str] = None,
    global_set: Optional[bool] = False
) -> Optional[str]:
  global _EXPERIMENT_PATH
  if experiment_path is None:
    if global_set:
      _EXPERIMENT_PATH = None
    return None
  experiment.validate_experiment_path(experiment_path)
  if global_set:
    _EXPERIMENT_PATH = experiment_path
  return experiment_path


def _set_logging_options(
    experiment_path: Optional[str] = None,
    logging_path: Optional[str] = None,
    logging_options: Optional[types.LoggingOptions] = None,
    global_set: Optional[bool] = False
) -> Tuple[types.LoggingOptions, Optional[str]]:
  if (
      logging_path is not None and
      logging_options is not None and
      logging_options.logging_path is not None):
    raise ValueError('logging_path and logging_options.logging_path are '
                     'both set. Either set logging_path or '
                     'logging_options.logging_path, but not both.')

  root_logging_path = None
  if logging_path:
    root_logging_path = logging_path
  elif logging_options and logging_options.logging_path:
    root_logging_path = logging_options.logging_path
  else:
    root_logging_path = None

  result_logging_options = types.LoggingOptions()
  if root_logging_path is not None:
    if not os.path.exists(root_logging_path):
      raise ValueError(
          f'Root logging path does not exist: {root_logging_path}')

    if experiment_path is not None:
      result_logging_options.logging_path = os.path.join(
          root_logging_path, experiment_path)
    else:
      result_logging_options.logging_path = root_logging_path
    if not os.path.exists(result_logging_options.logging_path):
      os.makedirs(result_logging_options.logging_path, exist_ok=True)
  else:
    result_logging_options.logging_path = None

  if logging_options is not None:
    result_logging_options.stdout = logging_options.stdout
    result_logging_options.hide_sensitive_content = (
        logging_options.hide_sensitive_content)

  if global_set:
    global _ROOT_LOGGING_PATH
    global _LOGGING_OPTIONS
    _ROOT_LOGGING_PATH = root_logging_path
    _LOGGING_OPTIONS = result_logging_options

  return (result_logging_options, root_logging_path)


def _set_cache_options(
    cache_path: Optional[str] = None,
    cache_options: Optional[types.CacheOptions] = None,
    global_set: Optional[bool] = False
) -> types.CacheOptions:
  if (
      cache_path is not None and
      cache_options is not None and
      cache_options.cache_path is not None):
    raise ValueError('cache_path and cache_options.cache_path are both set.'
                     'Either set cache_path or cache_options.cache_path, but '
                     'not both.')

  result_cache_options = types.CacheOptions()
  if cache_path:
    result_cache_options.cache_path = cache_path

  if cache_options:
    if cache_options.cache_path:
      result_cache_options.cache_path = cache_options.cache_path

    result_cache_options.unique_response_limit = (
        cache_options.unique_response_limit)
    result_cache_options.retry_if_error_cached = (
        cache_options.retry_if_error_cached)
    result_cache_options.clear_query_cache_on_connect = (
        cache_options.clear_query_cache_on_connect)

    result_cache_options.disable_model_cache = (
        cache_options.disable_model_cache)
    result_cache_options.clear_model_cache_on_connect = (
        cache_options.clear_model_cache_on_connect)
    result_cache_options.model_cache_duration = (
        cache_options.model_cache_duration)

  if global_set:
    global _CACHE_OPTIONS
    _CACHE_OPTIONS = result_cache_options
  return result_cache_options


def _set_proxdash_options(
    proxdash_options: Optional[types.ProxDashOptions] = None,
    global_set: Optional[bool] = False) -> types.ProxDashOptions:
  result_proxdash_options = types.ProxDashOptions()
  if proxdash_options is not None:
    result_proxdash_options.stdout = proxdash_options.stdout
    result_proxdash_options.hide_sensitive_content = (
        proxdash_options.hide_sensitive_content)
    result_proxdash_options.disable_proxdash = proxdash_options.disable_proxdash
    result_proxdash_options.api_key = proxdash_options.api_key
    result_proxdash_options.base_url = proxdash_options.base_url

  if global_set:
    global _PROXDASH_OPTIONS
    _PROXDASH_OPTIONS = result_proxdash_options
  return result_proxdash_options


def _set_allow_multiprocessing(
    allow_multiprocessing: Optional[bool] = None,
    global_set: Optional[bool] = False) -> Optional[bool]:
  if allow_multiprocessing is None:
    return None
  if global_set:
    global _ALLOW_MULTIPROCESSING
    _ALLOW_MULTIPROCESSING = allow_multiprocessing
  return allow_multiprocessing


def _set_model_test_timeout(
    model_test_timeout: Optional[int] = None,
    global_set: Optional[bool] = False) -> Optional[int]:
  if model_test_timeout is None:
    return None
  if model_test_timeout < 1:
    raise ValueError('model_test_timeout must be greater than 0.')
  if global_set:
    global _MODEL_TEST_TIMEOUT
    _MODEL_TEST_TIMEOUT = model_test_timeout
  return model_test_timeout


def _set_strict_feature_test(
    strict_feature_test: Optional[bool] = None,
    global_set: Optional[bool] = False) -> Optional[bool]:
  if strict_feature_test is None:
    return None
  if global_set:
    global _STRICT_FEATURE_TEST
    _STRICT_FEATURE_TEST = strict_feature_test
  return strict_feature_test


def _set_suppress_provider_errors(
    suppress_provider_errors: Optional[bool] = None,
    global_set: Optional[bool] = False) -> Optional[bool]:
  if suppress_provider_errors is None:
    return None
  if global_set:
    global _SUPPRESS_PROVIDER_ERRORS
    _SUPPRESS_PROVIDER_ERRORS = suppress_provider_errors
  return suppress_provider_errors


def _get_run_type() -> types.RunType:
  return _RUN_TYPE


def _get_hidden_run_key() -> str:
  return _HIDDEN_RUN_KEY


def _get_experiment_path() -> str:
  return _EXPERIMENT_PATH


def _get_logging_options() -> LoggingOptions:
  return _LOGGING_OPTIONS


def _get_cache_options() -> CacheOptions:
  return _CACHE_OPTIONS


def _get_proxdash_options() -> ProxDashOptions:
  return _PROXDASH_OPTIONS


def _get_model_connector(
    provider_model_identifier: types.ProviderModelIdentifierType
) -> model_connector.ProviderModelConnector:
  global _MODEL_CONNECTORS
  provider_model = model_configs.get_provider_model_config(
      provider_model_identifier)
  if provider_model in _MODEL_CONNECTORS:
    return _MODEL_CONNECTORS[provider_model]

  connector = model_registry.get_model_connector(provider_model)
  _MODEL_CONNECTORS[provider_model] = connector(
      get_run_type=_get_run_type,
      get_strict_feature_test=_get_strict_feature_test,
      get_query_cache_manager=_get_query_cache_manager,
      get_logging_options=_get_logging_options,
      get_proxdash_connection=_get_proxdash_connection,
      stats=_get_stats())
  return _MODEL_CONNECTORS[provider_model]


def _get_registered_model_connector(
    call_type: types.CallType = types.CallType.GENERATE_TEXT
) -> model_connector.ProviderModelConnector:
  global _REGISTERED_MODEL_CONNECTORS
  if call_type not in _REGISTERED_MODEL_CONNECTORS:
    if call_type == types.CallType.GENERATE_TEXT:
      available_models = get_available_models()
      if (
          not available_models.model_cache_manager or
          not available_models.model_cache_manager.get(
              types.CallType.GENERATE_TEXT).working_models):
        print('Checking available models, this may take a while...')
      models = get_available_models().list_models(return_all=True)
      for provider_model in model_configs.PREFERRED_DEFAULT_MODELS:
        if provider_model in models.working_models:
          _REGISTERED_MODEL_CONNECTORS[call_type] = _get_model_connector(
              provider_model)
          break
      if call_type not in _REGISTERED_MODEL_CONNECTORS:
        if models.working_models:
          _REGISTERED_MODEL_CONNECTORS[call_type] = _get_model_connector(
              models.working_models.pop())
        else:
          raise ValueError(
              'No working models found in current environment:\n'
              '* Please check your environment variables and try again.\n'
              '* You can use px.check_health() method as instructed in '
              'https://www.proxai.co/proxai-docs/check-health')
  return _REGISTERED_MODEL_CONNECTORS[call_type]


def _get_model_cache_manager() -> model_cache.ModelCacheManager:
  global _MODEL_CACHE_MANAGER
  if _MODEL_CACHE_MANAGER is None:
    _MODEL_CACHE_MANAGER = model_cache.ModelCacheManager(
        get_cache_options=_get_cache_options)
  if (_MODEL_CACHE_MANAGER.status !=
      types.ModelCacheManagerStatus.CACHE_PATH_NOT_FOUND):
    return _MODEL_CACHE_MANAGER

  if _DEFAULT_MODEL_CACHE_PATH is not None:
    return _DEFAULT_MODEL_CACHE_MANAGER

  raise ValueError('Model cache manager is not initialized and there is no '
                   'default model cache manager.')


def _get_query_cache_manager() -> query_cache.QueryCacheManager:
  global _QUERY_CACHE_MANAGER
  if _QUERY_CACHE_MANAGER is None:
    _QUERY_CACHE_MANAGER = query_cache.QueryCacheManager(
        get_cache_options=_get_cache_options)

  return _QUERY_CACHE_MANAGER


def _get_proxdash_connection() -> proxdash.ProxDashConnection:
  global _PROXDASH_CONNECTION
  if _PROXDASH_CONNECTION is None:
    _PROXDASH_CONNECTION = proxdash.ProxDashConnection(
        hidden_run_key=_get_hidden_run_key(),
        get_experiment_path=_get_experiment_path,
        get_logging_options=_get_logging_options,
        get_proxdash_options=_get_proxdash_options)
  return _PROXDASH_CONNECTION


def _get_allow_multiprocessing() -> bool:
  return _ALLOW_MULTIPROCESSING


def _get_model_test_timeout() -> int:
  return _MODEL_TEST_TIMEOUT


def _get_strict_feature_test() -> bool:
  return _STRICT_FEATURE_TEST


def _get_suppress_provider_errors() -> bool:
  return _SUPPRESS_PROVIDER_ERRORS


def _get_stats() -> Dict[stat_types.GlobalStatType, stat_types.RunStats]:
  return _STATS


def set_run_type(run_type: types.RunType):
  global _RUN_TYPE
  _RUN_TYPE = run_type


def set_model(
    provider_model: Optional[types.ProviderModelIdentifierType] = None,
    generate_text: Optional[types.ProviderModelIdentifierType] = None):
  global _REGISTERED_MODEL_CONNECTORS
  if provider_model and generate_text:
    raise ValueError('provider_model and generate_text cannot be set at the '
                     'same time. Please set one of them.')

  if provider_model is None and generate_text is None:
    raise ValueError('provider_model or generate_text must be set.')

  if generate_text:
    provider_model = generate_text

  type_utils.check_provider_model_identifier_type(provider_model)
  _REGISTERED_MODEL_CONNECTORS[
      types.CallType.GENERATE_TEXT] = _get_model_connector(provider_model)


def connect(
    experiment_path: Optional[str]=None,
    cache_path: Optional[str]=None,
    cache_options: Optional[CacheOptions]=None,
    logging_path: Optional[str]=None,
    logging_options: Optional[LoggingOptions]=None,
    proxdash_options: Optional[ProxDashOptions]=None,
    allow_multiprocessing: Optional[bool]=True,
    model_test_timeout: Optional[int]=25,
    strict_feature_test: Optional[bool]=False,
    suppress_provider_errors: Optional[bool]=False):
  _set_experiment_path(
      experiment_path=experiment_path,
      global_set=True)
  _set_logging_options(
      experiment_path=experiment_path,
      logging_path=logging_path,
      logging_options=logging_options,
      global_set=True)
  _set_cache_options(
      cache_path=cache_path,
      cache_options=cache_options,
      global_set=True)
  _set_proxdash_options(
      proxdash_options=proxdash_options,
      global_set=True)
  _set_allow_multiprocessing(
      allow_multiprocessing=allow_multiprocessing,
      global_set=True)
  _set_model_test_timeout(
      model_test_timeout=model_test_timeout,
      global_set=True)
  _set_strict_feature_test(
      strict_feature_test=strict_feature_test,
      global_set=True)
  _set_suppress_provider_errors(
      suppress_provider_errors=suppress_provider_errors,
      global_set=True)

  # This ensures updating model cache manager instead of default model cache
  # manager.
  if _MODEL_CACHE_MANAGER is not None:
    _MODEL_CACHE_MANAGER.apply_external_state_changes()

  model_cache_manager = _get_model_cache_manager()
  query_cache_manager = _get_query_cache_manager()
  proxdash_connection = _get_proxdash_connection()

  query_cache_manager.apply_external_state_changes()
  proxdash_connection.apply_external_state_changes()

  for model_connector in _MODEL_CONNECTORS.values():
    model_connector.apply_external_state_changes()

  cache_options = _get_cache_options()
  if cache_options.clear_model_cache_on_connect:
    model_cache_manager.clear_cache()
  if cache_options.clear_query_cache_on_connect:
    query_cache_manager.clear_cache()


def generate_text(
    prompt: Optional[str] = None,
    system: Optional[str] = None,
    messages: Optional[types.MessagesType] = None,
    max_tokens: Optional[int] = None,
    temperature: Optional[float] = None,
    stop: Optional[types.StopType] = None,
    provider_model: Optional[types.ProviderModelIdentifierType] = None,
    use_cache: Optional[bool] = None,
    unique_response_limit: Optional[int] = None,
    extensive_return: bool = False,
    suppress_provider_errors: Optional[bool] = None) -> Union[str, types.LoggingRecord]:
  if prompt is not None and messages is not None:
    raise ValueError('prompt and messages cannot be set at the same time.')
  if messages is not None:
    type_utils.check_messages_type(messages)

  if use_cache:
    query_cache_manager = _get_query_cache_manager()
    if query_cache_manager.status != types.QueryCacheManagerStatus.WORKING:
      raise ValueError(
          'use_cache is True but query cache is not working.\n'
          f'Query Cache Manager Status: {query_cache_manager.status}')
  elif use_cache is None:
    query_cache_manager = _get_query_cache_manager()
    use_cache = (
        query_cache_manager.status == types.QueryCacheManagerStatus.WORKING)

  if provider_model is not None:
    model_connector = _get_model_connector(
        provider_model_identifier=provider_model)
  else:
    model_connector = _get_registered_model_connector(
        call_type=types.CallType.GENERATE_TEXT)

  logging_record: types.LoggingRecord = model_connector.generate_text(
      prompt=prompt,
      system=system,
      messages=messages,
      max_tokens=max_tokens,
      temperature=temperature,
      stop=stop,
      use_cache=use_cache,
      unique_response_limit=unique_response_limit)
  if logging_record.response_record.error:
    if suppress_provider_errors or (
        suppress_provider_errors is None and _get_suppress_provider_errors()):
      if extensive_return:
        return logging_record
      return logging_record.response_record.error
    else:
      error_traceback = ''
      if logging_record.response_record.error_traceback:
        error_traceback = logging_record.response_record.error_traceback + '\n'
      raise Exception(error_traceback + logging_record.response_record.error)

  if extensive_return:
    return logging_record
  return logging_record.response_record.response


def get_summary(
    run_time: bool = False,
    json: bool = False) -> Union[stat_types.RunStats, Dict[str, Any]]:
  stat_value = None
  if run_time:
    stat_value = copy.deepcopy(_STATS[stat_types.GlobalStatType.RUN_TIME])
  else:
    stat_value = copy.deepcopy(_STATS[stat_types.GlobalStatType.SINCE_CONNECT])

  if json:
    return type_serializer.encode_run_stats(stat_value)

  class StatValue(stat_types.RunStats):
    def __init__(self, stat_value):
      super().__init__(**stat_value.__dict__)

    def serialize(self):
      return type_serializer.encode_run_stats(self)

  return StatValue(stat_value)


def get_available_models() -> available_models.AvailableModels:
  global _AVAILABLE_MODELS
  if _AVAILABLE_MODELS is None:
    _AVAILABLE_MODELS = available_models.AvailableModels(
        get_run_type=_get_run_type,
        get_model_connector=_get_model_connector,
        get_allow_multiprocessing=_get_allow_multiprocessing,
        get_model_test_timeout=_get_model_test_timeout,
        get_logging_options=_get_logging_options,
        get_model_cache_manager=_get_model_cache_manager,
        get_proxdash_connection=_get_proxdash_connection)
  return _AVAILABLE_MODELS


def get_current_options(
    json: bool = False) -> Union[types.RunOptions, Dict[str, Any]]:
  run_options = types.RunOptions(
      run_type=_get_run_type(),
      hidden_run_key=_get_hidden_run_key(),
      experiment_path=_get_experiment_path(),
      root_logging_path=_ROOT_LOGGING_PATH,
      default_model_cache_path=_DEFAULT_MODEL_CACHE_PATH,
      logging_options=_get_logging_options(),
      cache_options=_get_cache_options(),
      proxdash_options=_get_proxdash_options(),
      strict_feature_test=_get_strict_feature_test(),
      suppress_provider_errors=_get_suppress_provider_errors(),
      allow_multiprocessing=_get_allow_multiprocessing(),
      model_test_timeout=_get_model_test_timeout())
  if json:
    return type_serializer.encode_run_options(run_options=run_options)
  return run_options


def reset_platform_cache():
  if _PLATFORM_USED_FOR_DEFAULT_MODEL_CACHE and _DEFAULT_MODEL_CACHE_MANAGER:
    _DEFAULT_MODEL_CACHE_MANAGER.clear_cache()


def reset_state():
  reset_platform_cache()
  _init_globals()


def check_health(
    experiment_path: Optional[str]=None,
    verbose: bool = True,
    allow_multiprocessing: bool = True,
    model_test_timeout: int = 25,
    extensive_return: bool = False,
) -> types.ModelStatus:
  if experiment_path is None:
    if _get_experiment_path() is None:
      now = datetime.datetime.now()
      experiment_path = (
          f'connection_health/{now.strftime("%Y-%m-%d_%H-%M-%S")}')
      experiment_path = _set_experiment_path(experiment_path=experiment_path)
      logging_options, _ = _set_logging_options(
          experiment_path=experiment_path,
          logging_options=_get_logging_options())
    else:
      experiment_path = _get_experiment_path()
      logging_options = _get_logging_options()
  if _get_run_type() == types.RunType.TEST:
    proxdash_options = types.ProxDashOptions(
        stdout=False,
        disable_proxdash=True)
  else:
    proxdash_options = copy.deepcopy(_get_proxdash_options())
    proxdash_options.stdout = verbose

  proxdash_connection = proxdash.ProxDashConnection(
      hidden_run_key=_get_hidden_run_key(),
      experiment_path=experiment_path,
      logging_options=logging_options,
      proxdash_options=proxdash_options)

  def _get_modified_model_connector(
      provider_model_identifier: types.ProviderModelIdentifierType
  ) -> model_connector.ProviderModelConnector:
    provider_model = model_configs.get_provider_model_config(
        provider_model_identifier)
    connector = model_registry.get_model_connector(provider_model)
    return connector(
        get_run_type=_get_run_type,
        get_strict_feature_test=_get_strict_feature_test,
        get_query_cache_manager=_get_query_cache_manager,
        logging_options=logging_options,
        proxdash_connection=proxdash_connection,
        stats=_get_stats())

  allow_multiprocessing = _set_allow_multiprocessing(
      allow_multiprocessing=allow_multiprocessing)
  model_test_timeout = _set_model_test_timeout(
      model_test_timeout=model_test_timeout)
  if verbose:
    print('> Starting to test each model...')
  models = available_models.AvailableModels(
      run_type=_get_run_type(),
      logging_options=logging_options,
      proxdash_connection=proxdash_connection,
      allow_multiprocessing=allow_multiprocessing,
      model_test_timeout=model_test_timeout,
      get_model_connector=_get_modified_model_connector)
  model_status = models.list_models(
      verbose=verbose, return_all=True)
  if verbose:
    providers = set(
        [model.provider for model in model_status.working_models] +
        [model.provider for model in model_status.failed_models])
    result_table = {
        provider: {'working': [], 'failed': []} for provider in providers}
    for model in model_status.working_models:
      result_table[model.provider]['working'].append(model.model)
    for model in model_status.failed_models:
      result_table[model.provider]['failed'].append(model.model)
    print('> Finished testing.\n'
          f'   Registered Providers: {len(providers)}\n'
          f'   Succeeded Models: {len(model_status.working_models)}\n'
          f'   Failed Models: {len(model_status.failed_models)}')
    for provider in sorted(providers):
      print(f'> {provider}:')
      for model in sorted(result_table[provider]['working']):
        provider_model = model_configs.get_provider_model_config(
            (provider, model))
        duration = model_status.provider_queries[
            provider_model].response_record.response_time
        print(f'   [ WORKING | {duration.total_seconds():6.2f}s ]: {model}')
      for model in sorted(result_table[provider]['failed']):
        provider_model = model_configs.get_provider_model_config(
            (provider, model))
        duration = model_status.provider_queries[
            provider_model].response_record.response_time
        print(f'   [ FAILED  | {duration.total_seconds():6.2f}s ]: {model}')
  if proxdash_connection.status == types.ProxDashConnectionStatus.CONNECTED:
    logging_utils.log_proxdash_message(
        logging_options=logging_options,
        proxdash_options=proxdash_options,
        message='Results are uploaded to the ProxDash.',
        type=types.LoggingType.INFO)
  if extensive_return:
    return model_status



import functools
from typing import Callable
import proxai.connectors.model_connector as model_connector
import proxai.connectors.providers.openai as openai_provider
import proxai.connectors.providers.claude as claude_provider
import proxai.connectors.providers.gemini as gemini_provider
import proxai.connectors.providers.cohere_api as cohere_api_provider
import proxai.connectors.providers.databricks as databricks_provider
import proxai.connectors.providers.mistral as mistral_provider
import proxai.connectors.providers.huggingface as huggingface_provider
import proxai.connectors.providers.mock_provider as mock_provider
import proxai.connectors.providers.deepseek as deepseek_provider
import proxai.connectors.providers.grok as grok_provider
import proxai.types as types
import proxai.connectors.model_configs as model_configs

_MODEL_CONNECTOR_MAP = {
  'openai': openai_provider.OpenAIConnector,
  'claude': claude_provider.ClaudeConnector,
  'gemini': gemini_provider.GeminiConnector,
  'cohere': cohere_api_provider.CohereConnector,
  'databricks': databricks_provider.DatabricksConnector,
  'mistral': mistral_provider.MistralConnector,
  'huggingface': huggingface_provider.HuggingFaceConnector,
  'deepseek': deepseek_provider.DeepSeekConnector,
  'grok': grok_provider.GrokConnector,
  'mock_provider': mock_provider.MockProviderModelConnector,
  'mock_failing_provider': mock_provider.MockFailingProviderModelConnector,
  'mock_slow_provider': mock_provider.MockSlowProviderModelConnector,
}


def get_model_connector(
    provider_model_identifier: types.ProviderModelIdentifierType,
    without_additional_args: bool = False
) -> Callable[[], model_connector.ProviderModelConnector]:
  provider_model = model_configs.get_provider_model_config(
      provider_model_identifier)
  if provider_model.provider not in _MODEL_CONNECTOR_MAP:
    raise ValueError(f'Provider not supported. {provider_model.provider}')
  connector = _MODEL_CONNECTOR_MAP[provider_model.provider]
  if without_additional_args:
    return connector
  return functools.partial(connector, provider_model=provider_model)



import os
import copy
import functools
from databricks.sdk import WorkspaceClient
import proxai.types as types
import proxai.connectors.providers.databricks_mock as databricks_mock
import proxai.connectors.model_connector as model_connector


class DatabricksConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'databricks'

  def init_model(self):
    w = WorkspaceClient()
    return w.serving_endpoints.get_open_ai_client()

  def init_mock_model(self):
    return databricks_mock.DatabricksMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: Databricks tries to use same parameters with OpenAI.
    # Some parameters seems not working as expected for some models. For
    # example, the system instruction doesn't have any effect on the completion
    # for databricks-dbrx-instruct. But the stop parameter works as expected for
    # this model. However, system instruction works for
    # databricks-llama-2-70b-chat.
    query_messages = []
    if query_record.system != None:
      query_messages.append({'role': 'system', 'content': query_record.system})
    if query_record.prompt != None:
      query_messages.append({'role': 'user', 'content': query_record.prompt})
    if query_record.messages != None:
      query_messages.extend(query_record.messages)
    provider_model = query_record.provider_model

    create = functools.partial(
        self.api.chat.completions.create,
        model=provider_model.provider_model_identifier,
        messages=query_messages)
    if query_record.max_tokens != None:
      create = functools.partial(create, max_tokens=query_record.max_tokens)
    if query_record.temperature != None:
      create = functools.partial(create, temperature=query_record.temperature)
    if query_record.stop != None:
      create = functools.partial(create, stop=query_record.stop)

    completion = create()
    return completion.choices[0].message.content



import copy
import datetime
import multiprocessing
import os
import traceback
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import proxai.types as types
import proxai.caching.model_cache as model_cache
import proxai.connections.proxdash as proxdash
import proxai.connectors.model_registry as model_registry
import proxai.connectors.model_connector as model_connector
import proxai.connectors.model_configs as model_configs
import proxai.logging.utils as logging_utils
import proxai.state_controllers.state_controller as state_controller
import proxai.type_utils as type_utils
import concurrent.futures

_AVAILABLE_MODELS_STATE_PROPERTY = '_available_models_state'
_GENERATE_TEXT_TEST_PROMPT = 'Hello model!'
_GENERATE_TEXT_TEST_MAX_TOKENS = 1000


class AvailableModels(state_controller.StateControlled):
  _model_cache_manager: Optional[model_cache.ModelCacheManager]
  _run_type: types.RunType
  _get_run_type: Callable[[], types.RunType]
  _get_model_connector: Callable[
      [types.ProviderModelType], model_connector.ProviderModelConnector]
  _cache_options: types.CacheOptions
  _get_cache_options: Callable[[], types.CacheOptions]
  _logging_options: types.LoggingOptions
  _get_logging_options: Callable[[], types.LoggingOptions]
  _allow_multiprocessing: bool
  _get_allow_multiprocessing: Callable[[], bool]
  _model_test_timeout: int
  _get_model_test_timeout: Callable[[], int]
  _proxdash_connection: proxdash.ProxDashConnection
  _get_proxdash_connection: Callable[[], proxdash.ProxDashConnection]
  _providers_with_key: Set[str]
  _latest_model_cache_path_used_for_update: Optional[str]
  _available_models_state: types.AvailableModelsState

  def __init__(
      self,
      run_type: Optional[types.RunType] = None,
      get_run_type: Optional[Callable[[], types.RunType]] = None,
      model_cache_manager: Optional[model_cache.ModelCacheManager] = None,
      get_model_cache_manager: Optional[
          Callable[[], model_cache.ModelCacheManager]] = None,
      logging_options: Optional[types.LoggingOptions] = None,
      get_logging_options: Optional[Callable[[], types.LoggingOptions]] = None,
      proxdash_connection: Optional[proxdash.ProxDashConnection] = None,
      get_proxdash_connection: Optional[
          Callable[[], proxdash.ProxDashConnection]] = None,
      allow_multiprocessing: bool = None,
      get_allow_multiprocessing: Optional[Callable[[], bool]] = None,
      model_test_timeout: Optional[int] = None,
      get_model_test_timeout: Optional[Callable[[], int]] = None,
      get_model_connector: Optional[
          Callable[[], model_connector.ProviderModelConnector]] = None,
      init_state: Optional[types.AvailableModelsState] = None):

    if init_state and (
        run_type is not None or
        get_run_type is not None or
        model_cache_manager is not None or
        get_model_cache_manager is not None or
        logging_options is not None or
        get_logging_options is not None or
        proxdash_connection is not None or
        get_proxdash_connection is not None or
        allow_multiprocessing is not None or
        get_allow_multiprocessing is not None or
        model_test_timeout is not None or
        get_model_test_timeout is not None):
      raise ValueError(
          'init_state and other parameters cannot be set at the same time.')

    super().__init__(
        run_type=run_type,
        get_run_type=get_run_type,
        model_cache_manager=model_cache_manager,
        get_model_cache_manager=get_model_cache_manager,
        logging_options=logging_options,
        get_logging_options=get_logging_options,
        proxdash_connection=proxdash_connection,
        get_proxdash_connection=get_proxdash_connection,
        allow_multiprocessing=allow_multiprocessing,
        get_allow_multiprocessing=get_allow_multiprocessing,
        model_test_timeout=model_test_timeout,
        get_model_test_timeout=get_model_test_timeout)

    self.init_state()

    if init_state:
      self.load_state(init_state)
    else:
      initial_state = self.get_state()
      self._get_run_type = get_run_type
      self._get_model_cache_manager = get_model_cache_manager
      self._get_logging_options = get_logging_options
      self._get_proxdash_connection = get_proxdash_connection
      self._get_allow_multiprocessing = get_allow_multiprocessing
      self._get_model_test_timeout = get_model_test_timeout
      self._get_model_connector = get_model_connector

      self.run_type = run_type
      self.model_cache_manager = model_cache_manager
      self.logging_options = logging_options
      self.proxdash_connection = proxdash_connection
      self.allow_multiprocessing = allow_multiprocessing
      self.model_test_timeout = model_test_timeout
      self.get_model_connector = get_model_connector

      self.providers_with_key = set()
      self.latest_model_cache_path_used_for_update = None
      self._load_provider_keys()

      self.handle_changes(initial_state, self.get_state())

  def get_internal_state_property_name(self):
    return _AVAILABLE_MODELS_STATE_PROPERTY

  def get_internal_state_type(self):
    return types.AvailableModelsState

  def handle_changes(
      self,
      old_state: types.AvailableModelsState,
      current_state: types.AvailableModelsState):
    pass

  def _load_provider_keys(self):
    self.providers_with_key = set()
    for provider, provider_key_name in model_configs.PROVIDER_KEY_MAP.items():
      provider_flag = True
      for key_name in provider_key_name:
        if key_name not in os.environ:
          provider_flag = False
          break
      if provider_flag:
        self.providers_with_key.add(provider)

  @property
  def run_type(self) -> types.RunType:
    return self.get_property_value('run_type')

  @run_type.setter
  def run_type(self, run_type: types.RunType):
    self.set_property_value('run_type', run_type)

  @property
  def logging_options(self) -> types.LoggingOptions:
    return self.get_property_value('logging_options')

  @logging_options.setter
  def logging_options(self, logging_options: types.LoggingOptions):
    self.set_property_value('logging_options', logging_options)

  @property
  def model_cache_manager(self) -> model_cache.ModelCacheManager:
    return self.get_state_controlled_property_value('model_cache_manager')

  @model_cache_manager.setter
  def model_cache_manager(
      self, value: model_cache.ModelCacheManager):
    self.set_state_controlled_property_value('model_cache_manager', value)

  @property
  def proxdash_connection(self) -> proxdash.ProxDashConnection:
    return self.get_state_controlled_property_value('proxdash_connection')

  @proxdash_connection.setter
  def proxdash_connection(self, value: proxdash.ProxDashConnection):
    self.set_state_controlled_property_value('proxdash_connection', value)

  @property
  def allow_multiprocessing(self) -> bool:
    return self.get_property_value('allow_multiprocessing')

  @allow_multiprocessing.setter
  def allow_multiprocessing(self, value: bool):
    self.set_property_value('allow_multiprocessing', value)

  @property
  def model_test_timeout(self) -> int:
    return self.get_property_value('model_test_timeout')

  @model_test_timeout.setter
  def model_test_timeout(self, value: int):
    self.set_property_value('model_test_timeout', value)

  @property
  def providers_with_key(self) -> Set[str]:
    return self.get_property_value('providers_with_key')

  @providers_with_key.setter
  def providers_with_key(self, value: Set[str]):
    self.set_property_value('providers_with_key', value)

  @property
  def latest_model_cache_path_used_for_update(self) -> Optional[str]:
    return self.get_property_value('latest_model_cache_path_used_for_update')

  @latest_model_cache_path_used_for_update.setter
  def latest_model_cache_path_used_for_update(
      self, value: Optional[str]):
    self.set_property_value('latest_model_cache_path_used_for_update', value)

  def _get_all_models(self, models: types.ModelStatus, call_type: str):
    if call_type == types.CallType.GENERATE_TEXT:
      for provider_models in model_configs.GENERATE_TEXT_MODELS.values():
        for provider_model in provider_models.values():
          models.unprocessed_models.add(provider_model)

  def _filter_by_provider_api_key(self, models: types.ModelStatus):
    def _filter_set(
        provider_model_set: Set[types.ProviderModelType]
    ) -> Tuple[Set[types.ProviderModelType], Set[types.ProviderModelType]]:
      not_filtered_models = set()
      for provider_model in provider_model_set:
        if provider_model.provider not in self.providers_with_key:
          models.filtered_models.add(provider_model)
        else:
          not_filtered_models.add(provider_model)
      return not_filtered_models
    models.unprocessed_models = _filter_set(models.unprocessed_models)
    models.working_models = _filter_set(models.working_models)
    models.failed_models = _filter_set(models.failed_models)

  def _filter_by_providers(
      self,
      models: types.ModelStatus,
      providers: Optional[Set[str]] = None):
    if providers is None:
      return
    def _filter_set(
        provider_model_set: Set[types.ProviderModelType]
    ) -> Tuple[Set[types.ProviderModelType], Set[types.ProviderModelType]]:
      not_filtered_models = set()
      for provider_model in provider_model_set:
        if provider_model.provider not in providers:
          models.filtered_models.add(provider_model)
        else:
          not_filtered_models.add(provider_model)
      return not_filtered_models
    models.unprocessed_models = _filter_set(models.unprocessed_models)
    models.working_models = _filter_set(models.working_models)
    models.failed_models = _filter_set(models.failed_models)

  def _filter_by_provider_models(
      self,
      models: types.ModelStatus,
      provider_models: Optional[Set[types.ProviderModelType]] = None):
    if provider_models is None:
      return
    def _filter_set(
        provider_model_set: Set[types.ProviderModelType]
    ) -> Tuple[Set[types.ProviderModelType], Set[types.ProviderModelType]]:
      not_filtered_models = set()
      for provider_model in provider_model_set:
        if provider_model not in provider_models:
          models.filtered_models.add(provider_model)
        else:
          not_filtered_models.add(provider_model)
      return not_filtered_models
    models.unprocessed_models = _filter_set(models.unprocessed_models)
    models.working_models = _filter_set(models.working_models)
    models.failed_models = _filter_set(models.failed_models)

  def _filter_by_model_size(
      self,
      models: types.ModelStatus,
      model_size: Optional[types.ModelSizeType] = None):
    if model_size is None:
      return
    selected_model_config = None
    if model_size == types.ModelSizeType.SMALL:
      selected_model_config = model_configs.SMALL_GENERATE_TEXT_MODELS
    elif model_size == types.ModelSizeType.MEDIUM:
      selected_model_config = model_configs.MEDIUM_GENERATE_TEXT_MODELS
    elif model_size == types.ModelSizeType.LARGE:
      selected_model_config = model_configs.LARGE_GENERATE_TEXT_MODELS
    elif model_size == types.ModelSizeType.LARGEST:
      selected_model_config = model_configs.LARGEST_GENERATE_TEXT_MODELS

    allowed_models = set()
    for provider_models in selected_model_config.values():
      for provider_model in provider_models.values():
        allowed_models.add(provider_model)

    def _filter_set(
        provider_model_set: Set[types.ProviderModelType]
    ) -> Tuple[Set[types.ProviderModelType], Set[types.ProviderModelType]]:
      not_filtered_models = set()
      for provider_model in provider_model_set:
        if provider_model in allowed_models:
          not_filtered_models.add(provider_model)
        else:
          models.filtered_models.add(provider_model)
      return not_filtered_models

    models.unprocessed_models = _filter_set(models.unprocessed_models)
    models.working_models = _filter_set(models.working_models)
    models.failed_models = _filter_set(models.failed_models)

  def _filter_by_cache(
      self,
      models: types.ModelStatus,
      call_type: str):
    if not self.model_cache_manager:
      return
    cache_model_status = types.ModelStatus()
    if self.model_cache_manager:
      cache_model_status = self.model_cache_manager.get(call_type=call_type)
      self._place_filtered_models_back_to_working_or_failed_models(
          cache_model_status)

    for provider_model, query in cache_model_status.provider_queries.items():
      models.provider_queries[provider_model] = copy.deepcopy(query)

    for provider_model in list(models.unprocessed_models):
      if provider_model in cache_model_status.working_models:
        models.unprocessed_models.discard(provider_model)
        models.working_models.add(provider_model)
      elif provider_model in cache_model_status.failed_models:
        models.unprocessed_models.discard(provider_model)
        models.failed_models.add(provider_model)

  @staticmethod
  def _test_generate_text(
      provider_model_state: types.ProviderModelState,
      verbose: bool = False
  ) -> List[types.LoggingRecord]:
    if verbose:
      print(f'Testing {provider_model_state.provider_model}...')
    start_utc_date = datetime.datetime.now(datetime.timezone.utc)
    model_connector = model_registry.get_model_connector(
        provider_model_state.provider_model,
        without_additional_args=True)
    model_connector = model_connector(init_state=provider_model_state)
    try:
      logging_record: types.LoggingRecord = model_connector.generate_text(
          prompt=_GENERATE_TEXT_TEST_PROMPT,
          use_cache=False)
      return logging_record
    except Exception as e:
      return types.LoggingRecord(
          query_record=types.QueryRecord(
              call_type=types.CallType.GENERATE_TEXT,
              provider_model=provider_model_state.provider_model,
              prompt=_GENERATE_TEXT_TEST_PROMPT,
              max_tokens=_GENERATE_TEXT_TEST_MAX_TOKENS),
          response_record=types.QueryResponseRecord(
              error=str(e),
              error_traceback=traceback.format_exc(),
              start_utc_date=start_utc_date,
              end_utc_date=datetime.datetime.now(datetime.timezone.utc),
              response_time=(
                  datetime.datetime.now(datetime.timezone.utc)
                  - start_utc_date)),
          response_source=types.ResponseSource.PROVIDER)

  def _get_timeout_logging_record(
      self,
      provider_model: types.ProviderModelType):
    end_utc_date = datetime.datetime.now(datetime.timezone.utc)
    start_utc_date = end_utc_date - datetime.timedelta(
        seconds=self.model_test_timeout)
    return types.LoggingRecord(
        query_record=types.QueryRecord(
            call_type=types.CallType.GENERATE_TEXT,
            provider_model=provider_model,
            prompt=_GENERATE_TEXT_TEST_PROMPT,
            max_tokens=_GENERATE_TEXT_TEST_MAX_TOKENS),
        response_record=types.QueryResponseRecord(
            error=(
                f'Model {provider_model} took longer than '
                f'{self.model_test_timeout} seconds to respond'),
            error_traceback=traceback.format_exc(),
            start_utc_date=start_utc_date,
            end_utc_date=end_utc_date,
            response_time=(end_utc_date - start_utc_date)),
          response_source=types.ResponseSource.PROVIDER)

  def _get_bootstrap_error(self, e: Exception):
    error_str = str(e).lower()
    is_bootstrapping_error = (
      "an attempt has been made to start a new process before" in error_str and
      "current process has finished its bootstrapping phase" in error_str
    )
    if is_bootstrapping_error:
      return Exception(
          f'{e}\n\nMultiprocessing initialization error: Unable to start new '
          'processes because the proxai library was imported and used '
          'outside of the "if __name__ == \'__main__\':" block. To fix '
          'this:\n'
          '1. Move your proxai code inside a "if __name__ == \'__main__\':"'
          ' block, or\n'
          '2. Disable multiprocessing by setting '
          'allow_multiprocessing=False on px.connect()\n'
          'For more details, see followings:\n'
          '- https://www.proxai.co/proxai-docs/advanced/multiprocessing\n'
          '- https://docs.python.org/3/library/multiprocessing.html#the-'
          'spawn-and-forkserver-start-methods\n')
    return None


  def _test_models_with_multiprocessing(
      self,
      model_connectors: Dict[
        types.ProviderModelType, model_connector.ProviderModelConnector],
      call_type: str,
      verbose: bool = False):
    process_count = max(1, multiprocessing.cpu_count() - 1)
    test_func = None
    if call_type == types.CallType.GENERATE_TEXT:
      test_func = self._test_generate_text
    else:
      raise ValueError(f'Call type not supported: {call_type}')

    test_results = []
    try:
      pool = multiprocessing.Pool(processes=process_count)
      pool_results = []
      for provider_model, connector in model_connectors.items():
        pool_result = pool.apply_async(
            test_func,
            args=(connector.get_state(), verbose))
        pool_results.append((provider_model, pool_result))
      pool.close()
      for provider_model, pool_result in pool_results:
        try:
          test_results.append(pool_result.get(timeout=self.model_test_timeout))
        except multiprocessing.TimeoutError:
          if verbose:
            print(
                f"> {provider_model} query took longer than "
                f'{self.model_test_timeout} seconds to respond')
          test_results.append(self._get_timeout_logging_record(provider_model))
      pool.terminate()
      pool.join()
    except Exception as e:
      bootstrap_error = self._get_bootstrap_error(e)
      if bootstrap_error:
        raise bootstrap_error
      else:
        raise e
    return test_results

  def _test_models_sequentially(
      self,
      model_connectors: Dict[
        types.ProviderModelType, model_connector.ProviderModelConnector],
      call_type: str,
      verbose: bool = False):
    """Tests provider models sequentially.

    This function is used when multiprocessing is disabled. Note that this
    function cannot handle model timeouts because of the python's limitation
    on timeout handling without multiprocessing/threading.
    """
    test_func = None
    if call_type == types.CallType.GENERATE_TEXT:
      test_func = self._test_generate_text
    else:
      raise ValueError(f'Call type not supported: {call_type}')

    warning_message = (
      'Testing models sequentially can take a while because it '
      'is not possible to handle model timeouts without multiprocessing.\n'
      'Some models may take very long time to respond. '
      'To speed up the test, set allow_multiprocessing=True.')
    logging_utils.log_message(
        logging_options=self.logging_options,
        message=warning_message,
        type=types.LoggingType.WARNING)
    # Todo: After adding more stdout control on px.types.LoggingOptions,
    #       following can be removed.
    if verbose:
      print(f'WARNING: {warning_message}')

    test_results = []
    for connector in model_connectors.values():
      test_results.append(test_func(connector.get_state(), verbose))

    return test_results

  def _test_models(
      self,
      models: types.ModelStatus,
      call_type: str,
      verbose: bool = False):
    if not models.unprocessed_models:
      return

    model_connectors = {}
    for provider_model in models.unprocessed_models:
      model_connectors[provider_model] = self.get_model_connector(
          provider_model)

    if self.allow_multiprocessing:
      test_results = self._test_models_with_multiprocessing(
          model_connectors=model_connectors,
          call_type=call_type,
          verbose=verbose)
    else:
      test_results = self._test_models_sequentially(
          model_connectors=model_connectors,
          call_type=call_type,
          verbose=verbose)

    for logging_record in test_results:
      models.unprocessed_models.discard(
          logging_record.query_record.provider_model)
      if logging_record.response_record.response is not None:
        models.working_models.add(logging_record.query_record.provider_model)
      else:
        models.failed_models.add(logging_record.query_record.provider_model)
      models.provider_queries[
          logging_record.query_record.provider_model] = logging_record

  def _format_set(
      self,
      provider_model_set: Set[types.ProviderModelType]
  ) -> List[types.ProviderModelType]:
    return sorted(list(provider_model_set))

  def _place_filtered_models_back_to_working_or_failed_models(
      self,
      models: types.ModelStatus):
    for provider_model in list(models.filtered_models):
      if provider_model in models.provider_queries:
        if models.provider_queries[
            provider_model].response_record.response is not None:
          models.filtered_models.discard(provider_model)
          models.working_models.add(provider_model)
        elif models.provider_queries[
            provider_model].response_record.error is not None:
          models.filtered_models.discard(provider_model)
          models.failed_models.add(provider_model)

  def _fetch_all_models(
      self,
      selected_providers: Optional[Set[str]] = None,
      selected_provider_models: Optional[Set[types.ProviderModelType]] = None,
      model_size: Optional[types.ModelSizeType] = None,
      verbose: bool = False,
      clear_model_cache: bool = False,
      call_type: types.CallType = types.CallType.GENERATE_TEXT
  ) -> types.ModelStatus:
    if call_type != types.CallType.GENERATE_TEXT:
      raise ValueError(f'Call type not supported: {call_type}')

    if self.model_cache_manager and clear_model_cache:
      self.model_cache_manager.clear_cache()

    start_utc_date = datetime.datetime.now(datetime.timezone.utc)
    models = types.ModelStatus()
    self._load_provider_keys()
    self._get_all_models(models, call_type=call_type)
    self._filter_by_provider_api_key(models)
    self._filter_by_providers(models, providers=selected_providers)
    self._filter_by_provider_models(
        models, provider_models=selected_provider_models)
    self._filter_by_model_size(models, model_size=model_size)
    self._filter_by_cache(models, call_type=types.CallType.GENERATE_TEXT)

    print_flag = bool(verbose and models.unprocessed_models)
    verbose_print = print if print_flag else lambda *args, **kwargs: None

    verbose_print(f'From cache;\n'
                  f'  {len(models.working_models)} models are working.\n'
                  f'  {len(models.failed_models)} models are failed.')

    if models.unprocessed_models:
      verbose_print(f'Running test for {len(models.unprocessed_models)} models.')
      self._test_models(
          models,
          call_type=types.CallType.GENERATE_TEXT,
          verbose=verbose)
      verbose_print(f'After test;\n'
                    f'  {len(models.working_models)} models are working.\n'
                    f'  {len(models.failed_models)} models are failed.')

    if self.model_cache_manager:
      self.model_cache_manager.save(model_status=models, call_type=call_type)
      self.latest_model_cache_path_used_for_update = (
          self.model_cache_manager.cache_path)

    end_utc_date = datetime.datetime.now(datetime.timezone.utc)
    duration = (end_utc_date - start_utc_date).total_seconds()
    verbose_print(f'Test duration: {duration} seconds.')

    return models

  def _check_model_cache_path_same(self):
    model_cache_path = None
    if not self.model_cache_manager:
      model_cache_path = None
    else:
      model_cache_path = self.model_cache_manager.cache_path
    return model_cache_path == self.latest_model_cache_path_used_for_update

  def list_models(
      self,
      model_size: Optional[types.ModelSizeIdentifierType] = None,
      verbose: bool = False,
      return_all: bool = False,
      clear_model_cache: bool = False,
      call_type: types.CallType = types.CallType.GENERATE_TEXT
  ) -> Union[List[types.ProviderModelType], types.ModelStatus]:
    if call_type != types.CallType.GENERATE_TEXT:
      raise ValueError(f'Call type not supported: {call_type}')
    if model_size is not None:
      model_size = type_utils.check_model_size_identifier_type(model_size)

    model_status: Optional[types.ModelStatus] = None
    if not self.model_cache_manager:
      logging_utils.log_message(
          logging_options=self.logging_options,
          message='Model cache is not enabled. Fetching all models from '
          'providers. This is not ideal for performance.',
          type=types.LoggingType.WARNING)
      model_status = self._fetch_all_models(
          model_size=model_size,
          call_type=call_type,
          verbose=verbose)
    elif (
        clear_model_cache or
        not self._check_model_cache_path_same()):
      model_status = self._fetch_all_models(
          model_size=model_size,
          clear_model_cache=clear_model_cache,
          call_type=call_type,
          verbose=verbose)
    else:
      model_status = self._fetch_all_models(
          model_size=model_size,
          call_type=call_type,
          verbose=verbose)

    if return_all:
      return model_status
    return self._format_set(model_status.working_models)

  def list_providers(
      self,
      verbose: bool = False,
      clear_model_cache: bool = False,
      call_type: types.CallType = types.CallType.GENERATE_TEXT
  ) -> List[str]:
    if call_type != types.CallType.GENERATE_TEXT:
      raise ValueError(f'Call type not supported: {call_type}')

    providers_with_key: Optional[Set[str]] = None
    if not self.model_cache_manager:
      # For performance, we only load the provider keys if the model cache is
      # not enabled instead of fetching all models.
      self._load_provider_keys()
      providers_with_key = self.providers_with_key
    elif (
        clear_model_cache or
        not self._check_model_cache_path_same()):
      model_status = self._fetch_all_models(
          verbose=verbose,
          clear_model_cache=clear_model_cache,
          call_type=call_type)
      providers_with_key = set([
          model.provider
          for model in model_status.working_models])
    else:
      model_status = self._fetch_all_models(
          verbose=verbose,
          call_type=call_type)
      providers_with_key = set([
          model.provider
          for model in model_status.working_models])

    return sorted(list(providers_with_key))

  def list_provider_models(
      self,
      provider: str,
      model_size: Optional[types.ModelSizeIdentifierType] = None,
      verbose: bool = False,
      return_all: bool = False,
      clear_model_cache: bool = False,
      call_type: types.CallType = types.CallType.GENERATE_TEXT,
  ) -> Union[List[types.ProviderModelType], types.ModelStatus]:
    if call_type != types.CallType.GENERATE_TEXT:
      raise ValueError(f'Call type not supported: {call_type}')
    if model_size is not None:
      model_size = type_utils.check_model_size_identifier_type(model_size)
    if provider not in model_configs.GENERATE_TEXT_MODELS:
      raise ValueError(
          f'Provider not found in model_configs: {provider}\n'
          f'Available providers: {model_configs.GENERATE_TEXT_MODELS.keys()}')

    model_status: Optional[types.ModelStatus] = None
    if not self.model_cache_manager:
      self._load_provider_keys()
      if provider not in self.providers_with_key:
        raise ValueError(
            f'Provider key not found in environment variables for {provider}.\n'
            f'Required keys: {model_configs.PROVIDER_KEY_MAP[provider]}')
      model_status = types.ModelStatus()
      for model in model_configs.GENERATE_TEXT_MODELS[provider].values():
        model_status.working_models.add(model)
      self._filter_by_model_size(model_status, model_size=model_size)
    elif (
        clear_model_cache or
        not self._check_model_cache_path_same()):
      model_status = self._fetch_all_models(
          selected_providers=set([provider]),
          model_size=model_size,
          verbose=verbose,
          clear_model_cache=clear_model_cache,
          call_type=call_type)
    else:
      model_status = self._fetch_all_models(
          selected_providers=set([provider]),
          model_size=model_size,
          verbose=verbose,
          call_type=call_type)

    if return_all:
      return model_status
    return self._format_set(model_status.working_models)

  def get_model(
      self,
      provider: str,
      model: str,
      verbose: bool = False,
      clear_model_cache: bool = False,
      allow_non_working_model: bool = False,
      call_type: types.CallType = types.CallType.GENERATE_TEXT
  ) -> types.ProviderModelType:
    if call_type != types.CallType.GENERATE_TEXT:
      raise ValueError(f'Call type not supported: {call_type}')

    if provider not in model_configs.GENERATE_TEXT_MODELS:
      raise ValueError(
          f'Provider not found in model_configs: {provider}\n'
          f'Available providers: {model_configs.GENERATE_TEXT_MODELS.keys()}')

    if model not in model_configs.GENERATE_TEXT_MODELS[provider]:
      raise ValueError(
          f'Model not found in {provider} models: {model}\n'
          f'Available models: {model_configs.GENERATE_TEXT_MODELS[provider]}')

    provider_model = model_configs.get_provider_model_config((provider, model))

    model_status: Optional[types.ModelStatus] = None
    if not self.model_cache_manager:
      # For performance, we only load the provider keys if the model cache is
      # not enabled instead of fetching all models.
      self._load_provider_keys()
      if provider_model.provider not in self.providers_with_key:
        raise ValueError(
            'Provider key not found in environment variables for '
            f'{provider_model.provider}.\n'
            'Required keys: '
            f'{model_configs.PROVIDER_KEY_MAP[provider_model.provider]}')
      return provider_model
    elif (
        clear_model_cache or
        not self._check_model_cache_path_same()):
      model_status = self._fetch_all_models(
          selected_provider_models=set([provider_model]),
          verbose=verbose,
          clear_model_cache=clear_model_cache,
          call_type=call_type)
    else:
      model_status = self._fetch_all_models(
          selected_provider_models=set([provider_model]),
          verbose=verbose,
          call_type=call_type)

    if provider_model in model_status.working_models:
      return provider_model

    if allow_non_working_model:
      logging_utils.log_message(
          logging_options=self.logging_options,
          message=(
              'Provider model not found in working models: '
              f'({provider}, {model})\nLogging Record: '
              f'{model_status.provider_queries.get(provider_model, "")}'),
          type=types.LoggingType.WARNING)
      return model_configs.ALL_MODELS[provider][model]

    raise ValueError(
        f'Provider model not found in working models: ({provider}, {model})\n'
        + 'Logging Record: '
        + f'{model_status.provider_queries.get(provider_model, "")}')



import anthropic
import copy
import functools
import proxai.types as types
import proxai.connectors.providers.claude_mock as claude_mock
import proxai.connectors.model_connector as model_connector


class ClaudeConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'claude'

  def init_model(self):
    return anthropic.Anthropic()

  def init_mock_model(self):
    return claude_mock.ClaudeMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: Claude uses 'user' and 'assistant' as roles. 'system' is a
    # different parameter.
    query_messages = []
    if query_record.prompt is not None:
      query_messages.append({'role': 'user', 'content': query_record.prompt})
    if query_record.messages is not None:
      query_messages.extend(query_record.messages)
    provider_model = query_record.provider_model

    create = functools.partial(
        self.api.messages.create,
        model=provider_model.provider_model_identifier,
        messages=query_messages)
    if query_record.system is not None:
      create = functools.partial(create, system=query_record.system)
    if query_record.max_tokens is not None:
      create = functools.partial(create, max_tokens=query_record.max_tokens)
    else:
      # Note: Claude models require a max_tokens parameter.
      create = functools.partial(create, max_tokens=4096)
    if query_record.temperature is not None:
      create = functools.partial(create, temperature=query_record.temperature)
    if query_record.stop is not None:
      if isinstance(query_record.stop, str):
        create = functools.partial(create, stop_sequences=[query_record.stop])
      else:
        create = functools.partial(create, stop_sequences=query_record.stop)

    completion = create()
    return completion.content[0].text



from typing import Dict, List, Optional


class _MockResponse(object):
  text: str

  def __init__(self):
    self.text = 'mock response'


class CohereMock(object):
  def chat(
      self,
      message: str,
      model: str,
      preamble: Optional[str]=None,
      chat_history: Optional[List[Dict]]=None,
      max_tokens: Optional[int]=None,
      temperature: Optional[float]=None,
      stop_sequences: Optional[List[str]]=None) -> _MockResponse:
    return _MockResponse()



from typing import Dict, List, Optional


class _MockMessage(object):
  content: str


class _MockChoice(object):
  message: _MockMessage

  def __init__(self):
    self.message = _MockMessage()
    self.message.content = 'mock response'


class _MockResponse(object):
  choices: List[_MockChoice]

  def __init__(self):
    self.choices = [_MockChoice()]


class _MockCompletions(object):
  def create(
      self,
      model: str,
      messages: List[Dict],
      max_tokens: Optional[int]=None,
      temperature: Optional[float]=None,
      stop: Optional[List[str]]=None) -> _MockResponse:
    return _MockResponse()


class _MockChat(object):
  completions: _MockCompletions

  def __init__(self):
    self.completions = _MockCompletions()


class DatabricksMock(object):
  chat: _MockChat

  def __init__(self):
    self.chat = _MockChat()



import copy
import functools
from mistralai.client import MistralClient
from mistralai.models.chat_completion import ChatMessage
import proxai.types as types
import proxai.connectors.providers.mistral_mock as mistral_mock
import proxai.connectors.model_connector as model_connector


class MistralConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'mistral'

  def init_model(self):
    return MistralClient()

  def init_mock_model(self):
    return mistral_mock.MistralMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: Mistral uses 'system', 'user', and 'assistant' as roles.
    query_messages = []
    if query_record.system != None:
      query_messages.append(
          ChatMessage(role='system', content=query_record.system))
    if query_record.prompt != None:
      query_messages.append(
          ChatMessage(role='user', content=query_record.prompt))
    if query_record.messages != None:
      for message in query_record.messages:
        if message['role'] == 'user':
          query_messages.append(
              ChatMessage(role='user', content=message['content']))
        if message['role'] == 'assistant':
          query_messages.append(
              ChatMessage(role='assistant', content=message['content']))
    provider_model = query_record.provider_model

    create = functools.partial(
        self.api.chat,
        model=provider_model.provider_model_identifier,
        messages=query_messages)
    if query_record.max_tokens != None:
      create = functools.partial(create, max_tokens=query_record.max_tokens)
    if query_record.temperature != None:
      create = functools.partial(create, temperature=query_record.temperature)

    completion = create()
    return completion.choices[0].message.content



import time
import proxai.types as types
import proxai.connectors.model_connector as model_connector


class MockProviderModelConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return "mock_provider"

  def init_model(self):
    return None

  def init_mock_model(self):
    return None

  def generate_text_proc(self, query_record: types.QueryRecord):
    return "mock response"


class MockFailingProviderModelConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return "mock_failing_provider"

  def init_model(self):
    return None

  def init_mock_model(self):
    return None

  def generate_text_proc(self, query_record: types.QueryRecord):
    raise ValueError('Temp Error')


class MockSlowProviderModelConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return "mock_slow_provider"

  def init_model(self):
    return None

  def init_mock_model(self):
    return None

  def generate_text_proc(self, query_record: types.QueryRecord):
    time.sleep(120)
    return "mock response"



"""Base class for state controllers that manage the internal state of a class.

Example:
class UserStateController(StateController):
  @classmethod
  def get_internal_state_property_name(cls):
    return '_user_state'

class User:
  def __init__(self):
    self._user_state = {}  # Internal state storage
    self._name = None

  @property
  @UserStateController.getter
  def name(self):
    return self._name

  @name.setter
  @UserStateController.setter
  def name(self, value):
    self._name = value

# Usage
user = User()
user.name = "Alice"  # Sets name and automatically updates internal state
print(user._user_state)  # {'name': 'Alice'}

Note: Current implementation makes deep copies if the value is not a primitive
type. Further optimizations can be made in the future for specific use cases.
"""
import copy
from functools import wraps
from typing import List, Callable, Any, Optional
from abc import ABC, abstractmethod
import dataclasses
import proxai.types as types


class BaseStateControlled(ABC):
  def __init__(self, **kwargs):
    pass


class StateControlled(BaseStateControlled):
  def __init__(self, **kwargs):
    available_properties = set([
        field.name
        for field in dataclasses.fields(self.get_internal_state_type())
    ])
    for raw_property_name, property_value in kwargs.items():
      property_name, property_getter_name = None, None
      if raw_property_name.startswith('get_'):
        property_name = self.get_property_name_from_func_getter_name(
            raw_property_name)
        property_getter_name = raw_property_name
      else:
        property_name = raw_property_name
        property_getter_name = self.get_property_func_getter_name(
            raw_property_name)

      if property_value is None:
        continue

      if property_name not in available_properties:
        raise ValueError(
            f'Invalid property name or property getter name:\n'
            f'Property name: {property_name}\n'
            f'Property getter name: {property_getter_name}\n'
            f'Available properties: {available_properties}')

      if (property_name in kwargs and
          kwargs[property_name] is not None and
          property_getter_name in kwargs and
          kwargs[property_getter_name] is not None):
        raise ValueError(
            f'Only one of {property_name} or {property_getter_name} should be set '
            'while initializing the StateControlled object.')

  @abstractmethod
  def get_internal_state_property_name(self):
    raise NotImplementedError('Subclasses must implement this method')

  @abstractmethod
  def get_internal_state_type(self):
    raise NotImplementedError('Subclasses must implement this method')

  @abstractmethod
  def handle_changes(
      self,
      old_state: dataclasses.dataclass,
      current_state: dataclasses.dataclass):
    raise NotImplementedError(
        'Subclasses must implement this method before using it.')

  @staticmethod
  def get_property_internal_name(field: str) -> str:
    return f'_{field}'

  @staticmethod
  def get_property_func_internal_getter_name(field: str) -> str:
    return f'_get_{field}'

  @staticmethod
  def get_property_func_getter_name(field: str) -> str:
    return f'get_{field}'

  @staticmethod
  def get_state_controlled_deserializer_name(field: str) -> str:
    return f'{field}_deserializer'

  @staticmethod
  def get_property_name_from_func_getter_name(func_getter_name: str) -> str:
    if not func_getter_name.startswith('get_'):
      raise ValueError(
          f'Invalid function getter name: {func_getter_name}. '
          'It should start with "get_".')
    return func_getter_name[4:]

  def get_property_internal_value(self, property_name: str) -> Any:
    """Direct internal value getter."""
    return getattr(self, self.get_property_internal_name(property_name))

  def get_property_func_getter(self, property_name: str) -> Any:
    func = getattr(
        self,
        self.get_property_func_internal_getter_name(property_name),
        None)
    if func is None:
      return None
    return func

  def get_property_internal_value_or_from_getter(
      self, property_name: str) -> Any:
    """Getter for the properties that support both direct and getter values.

    Returns the property value if it is set directly, otherwise returns the
    value from the getter function. If the property is not set directly or by
    the getter function, returns None.
    """
    literal_value = self.get_property_internal_value(property_name)
    if literal_value is not None:
      return literal_value

    func = self.get_property_func_getter(property_name)
    if func is not None:
      return func()

    return None

  def get_property_internal_state_value(self, property_name: str) -> Any:
    """Direct internal state value getter."""
    return getattr(
        getattr(self, self.get_internal_state_property_name()),
        property_name,
        None)

  def set_property_internal_value(
      self,
      property_name: str,
      value: Any):
    """Direct internal value setter."""
    setattr(self, self.get_property_internal_name(property_name), value)

  def set_property_internal_state_value(
      self,
      property_name: str,
      value: Any):
    """Sets the property value directly in the internal state."""
    # Note: This is not efficient but safely copies the value to the internal
    # state. In the future, we can optimize this further for specific use cases.
    if value is None:
      copied_value = None
    elif isinstance(value, (str, int, float, bool, type(None))):
      copied_value = value
    else:
      copied_value = copy.deepcopy(value)
    setattr(
      getattr(self, self.get_internal_state_property_name()),
      property_name,
      copied_value)

  def get_property_value(self, property_name: str) -> Any:
    result = self.get_property_internal_value_or_from_getter(property_name)
    self.set_property_internal_state_value(property_name, result)
    return result

  def set_property_value(self, property_name: str, value: Any):
    self.set_property_internal_value(property_name, value)
    # Call actual getter to get the updated state value:
    updated_value = getattr(self, property_name)
    self.set_property_internal_state_value(property_name, updated_value)

  def get_state_controlled_property_value(self, property_name: str) -> Any:
    result = self.get_property_internal_value_or_from_getter(property_name)

    if result is None:
      self.set_property_internal_state_value(property_name, None)
      return None

    if not isinstance(result, BaseStateControlled):
      raise ValueError(
          f'Invalid property value. Expected a StateControlled object. '
          f'Got: {type(result)}')

    self.set_property_internal_state_value(property_name, result.get_state())
    return result

  def set_state_controlled_property_value(
      self,
      property_name: str,
      value: Any):
    if value is None:
      self.set_property_internal_value(property_name, None)
    elif isinstance(value, BaseStateControlled):
      self.set_property_internal_value(property_name, value)
    elif isinstance(value, types.StateContainer):
      value = getattr(
          self,
          self.get_state_controlled_deserializer_name(property_name))(value)
      self.set_property_internal_value(property_name, value)
    else:
      raise ValueError(
          f'Invalid property value. Expected a StateControlled object. '
          f'Got: {type(value)}')

    # Call actual getter to get the updated state value:
    updated_value = getattr(self, property_name)

    if updated_value is None:
      self.set_property_internal_state_value(property_name, None)
      return

    if not isinstance(updated_value, BaseStateControlled):
      raise ValueError(
          f'Invalid property value. Expected a StateControlled object. '
          f'Got: {type(updated_value)}')

    self.set_property_internal_state_value(
        property_name, updated_value.get_state())

  def set_property_value_without_triggering_getters(
      self,
      property_name: str,
      value: Any):
    self.set_property_internal_value(property_name, value)
    if isinstance(value, BaseStateControlled):
      value = value.get_state()
    self.set_property_internal_state_value(property_name, value)

  def init_state(self):
    setattr(
      self,
      self.get_internal_state_property_name(),
      self.get_internal_state_type()())

    for field in dataclasses.fields(self.get_internal_state_type()):
      self.set_property_value_without_triggering_getters(
          field.name, field.default)
      if hasattr(self, self.get_property_func_internal_getter_name(field.name)):
        setattr(
          self, self.get_property_func_internal_getter_name(field.name), None)

    return self.get_state()

  def get_state(self) -> Any:
    result = self.get_internal_state_type()()
    for field in dataclasses.fields(self.get_internal_state_type()):
      value = getattr(self, field.name, None)
      if isinstance(value, BaseStateControlled):
        value = value.get_state()
      if value is not None:
        setattr(result, field.name, value)
    return result

  def get_internal_state(self) -> Any:
    return copy.deepcopy(
        getattr(self, self.get_internal_state_property_name()))

  def get_external_state_changes(self) -> Any:
    internal_state = self.get_internal_state()
    changes = self.get_internal_state_type()()
    state_changed = False
    for field in dataclasses.fields(self.get_internal_state_type()):
      property_value = getattr(self, field.name)
      if getattr(internal_state, field.name, None) != property_value:
        setattr(changes, field.name, property_value)
        state_changed = True

    if not state_changed:
      return None
    return changes

  def load_state(self, state: Any):
    if type(state) != self.get_internal_state_type():
      raise ValueError(
          f'Invalid state type.\nExpected: {self.get_internal_state_type()}\n'
          f'Actual: {type(state)}')

    for field in dataclasses.fields(self.get_internal_state_type()):
      value = getattr(state, field.name, None)
      if value is None:
        continue
      if isinstance(value, types.StateContainer):
        value = getattr(
            self,
            self.get_state_controlled_deserializer_name(field.name))(value)
      self.set_property_value_without_triggering_getters(field.name, value)

  def apply_state_changes(self, changes: Optional[Any] = None):
    if changes is None:
      changes = self.get_internal_state_type()()

    if type(changes) != self.get_internal_state_type():
      raise ValueError(
          f'Invalid state type.\nExpected: {self.get_internal_state_type()}\n'
          f'Actual: {type(changes)}')

    old_state = self.get_internal_state()
    self.load_state(changes)
    self.handle_changes(old_state, self.get_internal_state())

  def apply_external_state_changes(self) -> Any:
    old_state = self.get_internal_state()
    state_changed = False
    for field in dataclasses.fields(self.get_internal_state_type()):
      property_value = getattr(self, field.name)
      if getattr(old_state, field.name, None) != property_value:
        state_changed = True
    if not state_changed:
      return None
    self.handle_changes(old_state, self.get_internal_state())
